# pylint: disable=too-many-lines
from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common import (
    const,
    reports,
)
from pcs.lib.commands import resource
from pcs.lib.errors import LibraryError

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import outdent

TIMEOUT = 10


def create(
    env,
    wait=False,
    disabled=False,
    meta_attributes=None,
    operation_list=None,
    allow_invalid_operation=False,
    agent_name="ocf:heartbeat:Dummy",
    allow_invalid_instance_attributes=False,
):
    return resource.create(
        env,
        "A",
        agent_name,
        operation_list=operation_list if operation_list else [],
        meta_attributes=meta_attributes if meta_attributes else {},
        instance_attributes={},
        wait=wait,
        ensure_disabled=disabled,
        allow_invalid_operation=allow_invalid_operation,
        allow_invalid_instance_attributes=allow_invalid_instance_attributes,
    )


def create_group(
    env,
    wait=TIMEOUT,
    disabled=False,
    meta_attributes=None,
    operation_list=None,
):
    return resource.create_in_group(
        env,
        "A",
        "ocf:heartbeat:Dummy",
        "G",
        operation_list=operation_list if operation_list else [],
        meta_attributes=meta_attributes if meta_attributes else {},
        instance_attributes={},
        wait=wait,
        ensure_disabled=disabled,
    )


def create_clone(
    env,
    wait=TIMEOUT,
    disabled=False,
    meta_attributes=None,
    clone_options=None,
    operation_list=None,
    clone_id=None,
):
    return resource.create_as_clone(
        env,
        "A",
        "ocf:heartbeat:Dummy",
        operation_list=operation_list if operation_list else [],
        meta_attributes=meta_attributes if meta_attributes else {},
        instance_attributes={},
        clone_meta_options=clone_options if clone_options else {},
        clone_id=clone_id,
        wait=wait,
        ensure_disabled=disabled,
    )


def create_bundle(
    env,
    wait=TIMEOUT,
    disabled=False,
    meta_attributes=None,
    allow_not_accessible_resource=False,
    operation_list=None,
):
    return resource.create_into_bundle(
        env,
        "A",
        "ocf:heartbeat:Dummy",
        operation_list=operation_list if operation_list else [],
        meta_attributes=meta_attributes if meta_attributes else {},
        instance_attributes={},
        bundle_id="B",
        wait=wait,
        ensure_disabled=disabled,
        allow_not_accessible_resource=allow_not_accessible_resource,
    )


wait_error_message = outdent(
    """\
    Pending actions:
            Action 39: stonith-vm-rhel72-1-reboot  on vm-rhel72-1
    Error performing operation: Timer expired
    """
).strip()


def fixture_cib_primitive_stateful(
    use_legacy_roles=False, include_reload=True, res_id="S1"
):
    promoted_role = const.PCMK_ROLE_PROMOTED_PRIMARY
    unpromoted_role = const.PCMK_ROLE_UNPROMOTED_PRIMARY
    if use_legacy_roles:
        promoted_role = const.PCMK_ROLE_PROMOTED_LEGACY
        unpromoted_role = const.PCMK_ROLE_UNPROMOTED_LEGACY
    agent_reload = ""
    if include_reload:
        agent_reload = """
                    <op id="S1-reload-agent-interval-0s" interval="0s"
                        name="reload-agent" timeout="10s"/>
        """
    return f"""
            <primitive class="ocf" id="{res_id}" provider="pacemaker" type="Stateful">
                <operations>
                    <op id="S1-demote-interval-0s" interval="0s" name="demote"
                        timeout="10s"/>
                    <op id="S1-monitor-interval-10s" interval="10s"
                        name="monitor" role="{promoted_role}" timeout="20s"/>
                    <op id="S1-monitor-interval-11s" interval="11s"
                        name="monitor" role="{unpromoted_role}" timeout="20s"/>
                    <op id="S1-notify-interval-0s" interval="0s" name="notify"
                        timeout="5s"/>
                    <op id="S1-promote-interval-0s" interval="0s"
                        name="promote" timeout="10s"/>
                    {agent_reload}
                    <op id="S1-start-interval-0s" interval="0s" name="start"
                        role="{promoted_role}"/>
                    <op id="S1-stop-interval-0s" interval="0s" name="stop"
                        role="{unpromoted_role}"/>
                  </operations>
            </primitive>
        """


def fixture_cib_resources_xml(resources):
    return f"""
        <resources>
            {resources}
        </resources>
    """


fixture_cib_resources_xml_primitive_simplest = """
    <resources>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <operations>
                <op id="A-migrate_from-interval-0s" interval="0s"
                    name="migrate_from" timeout="20"
                />
                <op id="A-migrate_to-interval-0s" interval="0s"
                    name="migrate_to" timeout="20"
                />
                <op id="A-monitor-interval-10" interval="10" name="monitor"
                    timeout="20"
                />
                <op id="A-reload-interval-0s" interval="0s" name="reload"
                    timeout="20"
                />
                <op id="A-start-interval-0s" interval="0s" name="start"
                    timeout="20"
                />
                <op id="A-stop-interval-0s" interval="0s" name="stop"
                    timeout="20"
                />
            </operations>
        </primitive>
    </resources>
"""

fixture_cib_resources_xml_simplest_disabled = """<resources>
    <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
        <meta_attributes id="A-meta_attributes">
            <nvpair id="A-meta_attributes-target-role" name="target-role"
                value="Stopped"
            />
        </meta_attributes>
        <operations>
            <op id="A-migrate_from-interval-0s" interval="0s"
                name="migrate_from" timeout="20"
            />
            <op id="A-migrate_to-interval-0s" interval="0s" name="migrate_to"
                timeout="20"
            />
            <op id="A-monitor-interval-10" interval="10" name="monitor"
                timeout="20"
            />
            <op id="A-reload-interval-0s" interval="0s" name="reload"
                timeout="20"
            />
            <op id="A-start-interval-0s" interval="0s" name="start"
                timeout="20"
            />
            <op id="A-stop-interval-0s" interval="0s" name="stop" timeout="20"/>
        </operations>
    </primitive>
</resources>"""

fixture_cib_resources_xml_group_simplest = """<resources>
    <group id="G">
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <operations>
                <op id="A-migrate_from-interval-0s" interval="0s"
                    name="migrate_from" timeout="20"
                />
                <op id="A-migrate_to-interval-0s" interval="0s"
                    name="migrate_to" timeout="20"
                />
                <op id="A-monitor-interval-10" interval="10" name="monitor"
                    timeout="20"
                />
                <op id="A-reload-interval-0s" interval="0s" name="reload"
                    timeout="20"
                />
                <op id="A-start-interval-0s" interval="0s" name="start"
                    timeout="20"
                />
                <op id="A-stop-interval-0s" interval="0s" name="stop"
                    timeout="20"
                />
            </operations>
        </primitive>
    </group>
</resources>"""


fixture_cib_resources_xml_group_simplest_disabled = """<resources>
    <group id="G">
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <meta_attributes id="A-meta_attributes">
                <nvpair id="A-meta_attributes-target-role" name="target-role"
                    value="Stopped"
                />
            </meta_attributes>
            <operations>
                <op id="A-migrate_from-interval-0s" interval="0s"
                    name="migrate_from" timeout="20"
                />
                <op id="A-migrate_to-interval-0s" interval="0s"
                    name="migrate_to" timeout="20"
                />
                <op id="A-monitor-interval-10" interval="10" name="monitor"
                    timeout="20"
                />
                <op id="A-reload-interval-0s" interval="0s" name="reload"
                    timeout="20"
                />
                <op id="A-start-interval-0s" interval="0s" name="start"
                    timeout="20"
                />
                <op id="A-stop-interval-0s" interval="0s" name="stop"
                    timeout="20"
                />
            </operations>
        </primitive>
    </group>
</resources>"""


fixture_cib_resources_xml_clone_simplest_template = """<resources>
    <clone id="{clone_id}">
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <operations>
                <op id="A-migrate_from-interval-0s" interval="0s"
                    name="migrate_from" timeout="20"
                />
                <op id="A-migrate_to-interval-0s" interval="0s"
                    name="migrate_to" timeout="20"
                />
                <op id="A-monitor-interval-10" interval="10" name="monitor"
                    timeout="20"
                />
                <op id="A-reload-interval-0s" interval="0s" name="reload"
                    timeout="20"
                />
                <op id="A-start-interval-0s" interval="0s" name="start"
                    timeout="20"
                />
                <op id="A-stop-interval-0s" interval="0s" name="stop"
                    timeout="20"
                />
            </operations>
        </primitive>
    </clone>
</resources>"""


fixture_cib_resources_xml_clone_simplest = (
    fixture_cib_resources_xml_clone_simplest_template.format(clone_id="A-clone")
)


fixture_cib_resources_xml_clone_custom_id = (
    fixture_cib_resources_xml_clone_simplest_template.format(
        clone_id="CustomCloneId"
    )
)


fixture_cib_resources_xml_clone_simplest_disabled = """<resources>
    <clone id="A-clone">
        <meta_attributes id="A-clone-meta_attributes">
            <nvpair id="A-clone-meta_attributes-target-role"
                name="target-role"
                value="Stopped"
            />
        </meta_attributes>
        <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
            <operations>
                <op id="A-migrate_from-interval-0s" interval="0s"
                    name="migrate_from" timeout="20"
                />
                <op id="A-migrate_to-interval-0s" interval="0s"
                    name="migrate_to" timeout="20"
                />
                <op id="A-monitor-interval-10" interval="10" name="monitor"
                    timeout="20"
                />
                <op id="A-reload-interval-0s" interval="0s" name="reload"
                    timeout="20"
                />
                <op id="A-start-interval-0s" interval="0s" name="start"
                    timeout="20"
                />
                <op id="A-stop-interval-0s" interval="0s" name="stop"
                    timeout="20"
                />
            </operations>
        </primitive>
    </clone>
</resources>"""


def fixture_state_resources_xml(
    role="Started", failed="false", node_name="node1"
):
    return """
        <resources>
            <resource
                id="A" resource_agent="ocf::heartbeat:Dummy"
                role="{role}" failed="{failed}"
            >
                <node name="{node_name}" id="1" cached="false"/>
            </resource>
        </resources>
        """.format(
        role=role,
        failed=failed,
        node_name=node_name,
    )


class CreateRolesNormalization(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def prepare(self, ocf_1_0=True, cib_support=True):
        agent_file_name = None
        if ocf_1_0:
            agent_file_name = (
                "resource_agent_ocf_pacemaker_stateful_ocf_1.0.xml"
            )
        cib_file = "cib-empty-3.5.xml"
        if cib_support:
            cib_file = "cib-empty-3.7.xml"
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Stateful",
            agent_filename=agent_file_name,
        )
        self.config.runner.cib.load(filename=cib_file)
        self.config.runner.pcmk.resource_agent_self_validation(
            {},
            output="",
            standard="ocf",
            provider="pacemaker",
            agent_type="Stateful",
        )

    def create(self, operation_list=None):
        resource.create(
            self.env_assist.get_env(),
            "S1",
            "ocf:pacemaker:Stateful",
            operation_list=operation_list if operation_list else [],
            meta_attributes={},
            instance_attributes={},
        )

    def test_roles_normalization_user_defined(self):
        self.prepare(True, False)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(
                    use_legacy_roles=True, include_reload=False
                )
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED_LEGACY),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED_LEGACY),
            ]
        )

    def test_roles_normalization_user_defined_new_roles(self):
        self.prepare(True, False)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(
                    use_legacy_roles=True, include_reload=False
                )
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED),
            ]
        )

    def test_roles_normalization_user_defined_with_cib_support(self):
        self.prepare(True, True)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(include_reload=False)
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED_LEGACY),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED_LEGACY),
            ]
        )

    def test_roles_normalization_user_defined_new_roles_with_cib_support(self):
        self.prepare(True, True)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(include_reload=False)
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED),
            ]
        )

    def test_roles_normalization_agent(self):
        self.prepare(False, False)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(use_legacy_roles=True)
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED_LEGACY),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED_LEGACY),
            ]
        )

    def test_roles_normalization_agent_new_roles(self):
        self.prepare(False, False)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful(use_legacy_roles=True)
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED),
            ]
        )

    def test_roles_normalization_agent_with_cib_support(self):
        self.prepare(False, True)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful()
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED_LEGACY),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED_LEGACY),
            ]
        )

    def test_roles_normalization_agent_new_roles_with_cib_support(self):
        self.prepare(False, True)
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml(
                fixture_cib_primitive_stateful()
            )
        )
        self.create(
            [
                dict(name="start", role=const.PCMK_ROLE_PROMOTED),
                dict(name="stop", role=const.PCMK_ROLE_UNPROMOTED),
            ]
        )


class Create(TestCase):
    fixture_sanitized_operation = """
        <resources>
            <primitive class="ocf" id="A" provider="heartbeat"
                type="Dummy"
            >
                <operations>
                    <op id="A-migrate_from-interval-0s" interval="0s"
                        name="migrate_from" timeout="20"
                    />
                    <op id="A-migrate_to-interval-0s" interval="0s"
                        name="migrate_to" timeout="20"
                    />
                    <op id="A-monitor-interval-20" interval="20"
                        name="moni*tor" timeout="20"
                    />
                    <op id="A-monitor-interval-10" interval="10"
                        name="monitor" timeout="20"
                    />
                    <op id="A-reload-interval-0s" interval="0s"
                        name="reload" timeout="20"
                    />
                    <op id="A-start-interval-0s" interval="0s"
                        name="start" timeout="20"
                    />
                    <op id="A-stop-interval-0s" interval="0s"
                        name="stop" timeout="20"
                    />
                </operations>
            </primitive>
        </resources>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_simplest_resource(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_primitive_simplest
        )
        create(self.env_assist.get_env())

    def test_resource_self_validation_failure(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation(
            {},
            output="""
            <output source="stderr">not ignored</output>
            <output source="stdout">this is ignored</output>
            <output source="stderr">
            first issue
            another one
            </output>
            """,
            returncode=1,
        )
        self.env_assist.assert_raise_library_error(
            lambda: create(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.AGENT_SELF_VALIDATION_RESULT,
                    result="not ignored\nfirst issue\nanother one",
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_resource_self_validation_failure_forced(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation(
            {},
            output="""
            <output source="stderr">not ignored</output>
            <output source="stdout">this is ignored</output>
            <output source="stderr">
            first issue
            another one
            </output>
            """,
            returncode=1,
        )
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_primitive_simplest
        )
        create(
            self.env_assist.get_env(), allow_invalid_instance_attributes=True
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.AGENT_SELF_VALIDATION_RESULT,
                    result="not ignored\nfirst issue\nanother one",
                )
            ]
        )

    def test_resource_self_validation_invalid_output(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation(
            {},
            output="""<not valid> xml""",
            returncode=0,
        )
        self.env_assist.assert_raise_library_error(
            lambda: create(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.AGENT_SELF_VALIDATION_INVALID_DATA,
                    reason="Specification mandates value for attribute valid, line 7, column 29 (<string>, line 7)",
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_invalid_agent_name(self):
        self.env_assist.assert_raise_library_error(
            lambda: create(
                self.env_assist.get_env(),
                agent_name="ocf:heartbeat:something:else",
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_RESOURCE_AGENT_NAME,
                    name="ocf:heartbeat:something:else",
                )
            ]
        )

    def test_agent_guess_success(self):
        self.config.runner.pcmk.list_agents_standards("\n".join(["ocf"]))
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["agent1", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_pacemaker",
        )
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_primitive_simplest
        )
        create(self.env_assist.get_env(), agent_name="dummy")
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AGENT_NAME_GUESSED,
                    entered_name="dummy",
                    guessed_name="ocf:heartbeat:Dummy",
                ),
            ]
        )

    def test_agent_guess_ambiguous(self):
        self.config.runner.pcmk.list_agents_standards("\n".join(["ocf"]))
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_pacemaker",
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.create(
                self.env_assist.get_env(),
                "A",
                "dummy",
                [],
                {},
                {},
                allow_absent_agent=True,
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.AGENT_NAME_GUESS_FOUND_MORE_THAN_ONE,
                    agent="dummy",
                    possible_agents=[
                        "ocf:heartbeat:Dummy",
                        "ocf:pacemaker:Dummy",
                    ],
                )
            ],
        )

    def test_agent_guess_not_found(self):
        self.config.runner.pcmk.list_agents_standards("\n".join(["ocf"]))
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["agent1", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["agent1", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_pacemaker",
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.create(
                self.env_assist.get_env(),
                "A",
                "dummy",
                [],
                {},
                {},
                allow_absent_agent=True,
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.AGENT_NAME_GUESS_FOUND_NONE,
                    agent="dummy",
                )
            ],
        )

    def test_agent_load_failure(self):
        self.config.runner.pcmk.load_agent(
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.env_assist.assert_raise_library_error(
            lambda: create(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                    force_code=reports.codes.FORCE,
                    agent="ocf:heartbeat:Dummy",
                    reason=(
                        "Agent ocf:heartbeat:Dummy not found or does not support "
                        "meta-data: Invalid argument (22)\nMetadata query for "
                        "ocf:heartbeat:Dummy failed: Input/output error"
                    ),
                )
            ]
        )

    def test_agent_load_failure_forced(self):
        instance_attributes = {"parameters": "can", "be": "anything"}
        self.config.runner.pcmk.load_agent(
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.config.runner.cib.load()
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="ocf" id="C" provider="heartbeat" type="Dummy">
                        <instance_attributes id="C-instance_attributes">
                            <nvpair id="C-instance_attributes-be"
                                name="be" value="anything"/>
                            <nvpair id="C-instance_attributes-parameters"
                                name="parameters" value="can"/>
                        </instance_attributes>
                        <operations>
                            <op id="C-monitor-interval-60s"
                                interval="60s" name="monitor"/>
                        </operations>
                    </primitive>
                </resources>
            """
        )
        resource.create(
            self.env_assist.get_env(),
            "C",
            "ocf:heartbeat:Dummy",
            operation_list=[],
            meta_attributes={},
            instance_attributes=instance_attributes,
            allow_absent_agent=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="ocf:heartbeat:Dummy",
                    reason=(
                        "Agent ocf:heartbeat:Dummy not found or does not support "
                        "meta-data: Invalid argument (22)\nMetadata query for "
                        "ocf:heartbeat:Dummy failed: Input/output error"
                    ),
                )
            ]
        )

    def test_resource_with_operation(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="ocf" id="A" provider="heartbeat"
                        type="Dummy"
                    >
                        <operations>
                            <op id="A-migrate_from-interval-0s" interval="0s"
                                name="migrate_from" timeout="20"
                            />
                            <op id="A-migrate_to-interval-0s" interval="0s"
                                name="migrate_to" timeout="20"
                            />
                            <op id="A-monitor-interval-10" interval="10"
                                name="monitor" timeout="10s"
                            />
                            <op id="A-reload-interval-0s" interval="0s"
                                name="reload" timeout="20"
                            />
                            <op id="A-start-interval-0s" interval="0s"
                                name="start" timeout="20"
                            />
                            <op id="A-stop-interval-0s" interval="0s"
                                name="stop" timeout="20"
                            />
                        </operations>
                    </primitive>
                </resources>
            """
        )

        create(
            self.env_assist.get_env(),
            operation_list=[
                {"name": "monitor", "timeout": "10s", "interval": "10"}
            ],
        )

    def test_sanitize_operation_id_from_agent(self):
        self.config.runner.pcmk.load_agent(
            agent_filename=(
                "resource_agent_ocf_heartbeat_dummy_insane_action.xml"
            ),
        )
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(resources=self.fixture_sanitized_operation)
        create(self.env_assist.get_env())

    def test_sanitize_operation_id_from_user(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(resources=self.fixture_sanitized_operation)
        create(
            self.env_assist.get_env(),
            operation_list=[
                {"name": "moni*tor", "timeout": "20", "interval": "20"}
            ],
            allow_invalid_operation=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="operation name",
                    option_value="moni*tor",
                    allowed_values=[
                        "start",
                        "stop",
                        "monitor",
                        "reload",
                        "migrate_to",
                        "migrate_from",
                        "meta-data",
                        "validate-all",
                    ],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_resource_with_operation_depth(self):
        self.config.runner.pcmk.load_agent(
            stdout="""
                <resource-agent name="Dummy">
                    <version>1.0</version>
                    <shortdesc>Example stateless resource agent</shortdesc>
                    <parameters/>
                    <actions>
                        <action
                            name="monitor" timeout="20" interval="10" depth="0"
                        />
                        <action
                            name="monitor" timeout="20" interval="30" depth="10"
                        />
                    </actions>
                </resource-agent>
            """
        )
        self.config.runner.cib.load()
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="ocf" id="A" provider="heartbeat" type="Dummy">
                        <operations>
                            <op id="A-monitor-interval-10"
                                interval="10" name="monitor" timeout="20"
                            />
                            <op id="A-monitor-interval-30"
                                interval="30" name="monitor" timeout="20"
                            >
                                <instance_attributes
                                    id="A-monitor-interval-30-instance_attributes"
                                >
                                    <nvpair
                                        id="A-monitor-interval-30-instance_attributes-OCF_CHECK_LEVEL"
                                        name="OCF_CHECK_LEVEL" value="10"
                                    />
                                </instance_attributes>
                            </op>
                        </operations>
                    </primitive>
                </resources>
            """
        )
        create(self.env_assist.get_env())

    def test_unique_option(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive class="ocf" id="X" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="X-instance_attributes">
                            <nvpair
                                id="X-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="A" provider="pacemaker"
                        type="Dummy"
                    >
                        <instance_attributes id="A-instance_attributes">
                            <nvpair
                                id="A-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="B" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="B-instance_attributes">
                            <nvpair
                                id="B-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                </resources>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.create(
                self.env_assist.get_env(),
                "C",
                "ocf:heartbeat:Dummy",
                operation_list=[],
                meta_attributes={},
                instance_attributes={"state": "1"},
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE,
                    instance_attr_name="state",
                    instance_attr_value="1",
                    agent_name="ocf:heartbeat:Dummy",
                    resource_id_list=["B", "X"],
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_unique_option_forced(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive class="ocf" id="X" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="X-instance_attributes">
                            <nvpair
                                id="X-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="A" provider="pacemaker"
                        type="Dummy"
                    >
                        <instance_attributes id="A-instance_attributes">
                            <nvpair
                                id="A-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="B" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="B-instance_attributes">
                            <nvpair
                                id="B-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                </resources>
            """,
        )
        self.config.runner.pcmk.resource_agent_self_validation(
            dict(state=1), output=""
        )
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="ocf" id="X" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="X-instance_attributes">
                            <nvpair
                                id="X-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="A" provider="pacemaker"
                        type="Dummy"
                    >
                        <instance_attributes id="A-instance_attributes">
                            <nvpair
                                id="A-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="B" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="B-instance_attributes">
                            <nvpair
                                id="B-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                    </primitive>
                    <primitive class="ocf" id="C" provider="heartbeat"
                        type="Dummy"
                    >
                        <instance_attributes id="C-instance_attributes">
                            <nvpair
                                id="C-instance_attributes-state"
                                name="state"
                                value="1"
                            />
                        </instance_attributes>
                        <operations>
                            <op id="C-monitor-interval-10" interval="10"
                                name="monitor" timeout="20"
                            />
                        </operations>
                    </primitive>
                </resources>
            """
        )
        resource.create(
            self.env_assist.get_env(),
            "C",
            "ocf:heartbeat:Dummy",
            operation_list=[],
            meta_attributes={},
            instance_attributes={"state": "1"},
            use_default_operations=False,
            allow_invalid_instance_attributes=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE,
                    instance_attr_name="state",
                    instance_attr_value="1",
                    agent_name="ocf:heartbeat:Dummy",
                    resource_id_list=["B", "X"],
                )
            ]
        )

    def test_cib_upgrade_on_onfail_demote(self):
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load(
            filename="cib-empty-3.3.xml",
            name="load_cib_old_version",
        )
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(filename="cib-empty-3.4.xml")
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources="""
                <resources>
                    <primitive class="ocf" id="A" provider="heartbeat"
                        type="Dummy"
                    >
                        <operations>
                            <op id="A-migrate_from-interval-0s" interval="0s"
                                name="migrate_from" timeout="20"
                            />
                            <op id="A-migrate_to-interval-0s" interval="0s"
                                name="migrate_to" timeout="20"
                            />
                            <op id="A-monitor-interval-10" interval="10"
                                name="monitor" timeout="10" on-fail="demote"
                            />
                            <op id="A-reload-interval-0s" interval="0s"
                                name="reload" timeout="20"
                            />
                            <op id="A-start-interval-0s" interval="0s"
                                name="start" timeout="20"
                            />
                            <op id="A-stop-interval-0s" interval="0s"
                                name="stop" timeout="20"
                            />
                        </operations>
                    </primitive>
                </resources>
            """
        )

        create(
            self.env_assist.get_env(),
            operation_list=[
                {
                    "name": "monitor",
                    "timeout": "10",
                    "interval": "10",
                    "on-fail": "demote",
                }
            ],
        )
        self.env_assist.assert_reports(
            [fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL)]
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_api_rng/api-result.rng")
)
class CreateWait(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_primitive_simplest,
            wait=TIMEOUT,
        )

    def test_fail_wait(self):
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_primitive_simplest,
            wait=TIMEOUT,
            exception=LibraryError(
                reports.item.ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(wait_error_message)
                )
            ),
            instead="env.push_cib",
        )
        self.env_assist.assert_raise_library_error(
            lambda: create(self.env_assist.get_env(), wait=TIMEOUT),
            [fixture.report_wait_for_idle_timed_out(wait_error_message)],
            expected_in_processor=False,
        )

    def test_wait_ok_run_fail(self):
        self.config.runner.pcmk.load_state(
            resources=fixture_state_resources_xml(failed="true")
        )

        self.env_assist.assert_raise_library_error(
            lambda: create(self.env_assist.get_env(), wait=TIMEOUT)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    def test_wait_ok_run_ok(self):
        self.config.runner.pcmk.load_state(
            resources=fixture_state_resources_xml()
        )
        create(self.env_assist.get_env(), wait=TIMEOUT)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                ),
            ]
        )

    def test_wait_ok_disable_fail(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_state_resources_xml()
            ).env.push_cib(
                resources=fixture_cib_resources_xml_simplest_disabled,
                wait=TIMEOUT,
                instead="env.push_cib",
            )
        )

        self.env_assist.assert_raise_library_error(
            lambda: create(
                self.env_assist.get_env(), wait=TIMEOUT, disabled=True
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                ),
            ]
        )

    def test_wait_ok_disable_ok(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            ).env.push_cib(
                resources=fixture_cib_resources_xml_simplest_disabled,
                wait=TIMEOUT,
                instead="env.push_cib",
            )
        )

        create(self.env_assist.get_env(), wait=TIMEOUT, disabled=True)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    def test_wait_ok_disable_ok_by_target_role(self):
        (
            self.config.runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            ).env.push_cib(
                resources=fixture_cib_resources_xml_simplest_disabled,
                wait=TIMEOUT,
                instead="env.push_cib",
            )
        )
        create(
            self.env_assist.get_env(),
            wait=TIMEOUT,
            meta_attributes={"target-role": "Stopped"},
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )


class CreateInGroup(TestCase):
    def setUp(self):
        self.agent_self_validation_call_name = (
            "runner.pcmk.resource_agent_self_validation"
        )
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation(
            {}, output="", name=self.agent_self_validation_call_name
        )

    def test_simplest_resource(self):
        (
            self.config.env.push_cib(
                resources="""
                    <resources>
                        <group id="G">
                            <primitive class="ocf" id="A" provider="heartbeat"
                                type="Dummy"
                            >
                                <operations>
                                    <op id="A-migrate_from-interval-0s"
                                        interval="0s" name="migrate_from"
                                        timeout="20"
                                    />
                                    <op id="A-migrate_to-interval-0s"
                                        interval="0s" name="migrate_to"
                                        timeout="20"
                                    />
                                    <op id="A-monitor-interval-10" interval="10"
                                        name="monitor" timeout="20"
                                    />
                                    <op id="A-reload-interval-0s" interval="0s"
                                        name="reload" timeout="20"
                                    />
                                    <op id="A-start-interval-0s" interval="0s"
                                        name="start" timeout="20"
                                    />
                                    <op id="A-stop-interval-0s" interval="0s"
                                        name="stop" timeout="20"
                                    />
                                </operations>
                            </primitive>
                        </group>
                    </resources>
                """
            )
        )

        create_group(self.env_assist.get_env(), wait=False)

    def test_cib_upgrade_on_onfail_demote(self):
        self.config.remove(self.agent_self_validation_call_name)
        self.config.runner.cib.load(
            filename="cib-empty-3.3.xml",
            instead="runner.cib.load",
            name="load_cib_old_version",
        )
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(filename="cib-empty-3.4.xml")
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources="""
                <resources>
                    <group id="G">
                        <primitive class="ocf" id="A" provider="heartbeat"
                            type="Dummy"
                        >
                            <operations>
                                <op id="A-migrate_from-interval-0s"
                                    interval="0s" name="migrate_from"
                                    timeout="20"
                                />
                                <op id="A-migrate_to-interval-0s"
                                    interval="0s" name="migrate_to"
                                    timeout="20"
                                />
                                <op id="A-monitor-interval-10" interval="10"
                                    name="monitor" timeout="10" on-fail="demote"
                                />
                                <op id="A-reload-interval-0s" interval="0s"
                                    name="reload" timeout="20"
                                />
                                <op id="A-start-interval-0s" interval="0s"
                                    name="start" timeout="20"
                                />
                                <op id="A-stop-interval-0s" interval="0s"
                                    name="stop" timeout="20"
                                />
                            </operations>
                        </primitive>
                    </group>
                </resources>
            """
        )

        create_group(
            self.env_assist.get_env(),
            operation_list=[
                {
                    "name": "monitor",
                    "timeout": "10",
                    "interval": "10",
                    "on-fail": "demote",
                }
            ],
            wait=False,
        )
        self.env_assist.assert_reports(
            [fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_fail_wait(self):
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_group_simplest,
            wait=TIMEOUT,
            exception=LibraryError(
                reports.item.ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(wait_error_message)
                )
            ),
        )

        self.env_assist.assert_raise_library_error(
            lambda: create_group(self.env_assist.get_env()),
            [fixture.report_wait_for_idle_timed_out(wait_error_message)],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_fail(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_group_simplest, wait=TIMEOUT
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(failed="true")
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_group(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_DOES_NOT_RUN, resource_id="A"
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_ok(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_group_simplest, wait=TIMEOUT
            ).runner.pcmk.load_state(resources=fixture_state_resources_xml())
        )
        create_group(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_fail(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_group_simplest_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(resources=fixture_state_resources_xml())
        )

        self.env_assist.assert_raise_library_error(
            lambda: create_group(self.env_assist.get_env(), disabled=True)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_group_simplest_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_group(self.env_assist.get_env(), disabled=True)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok_by_target_role(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_group_simplest_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_group(
            self.env_assist.get_env(),
            meta_attributes={"target-role": "Stopped"},
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )


class CreateAsClone(TestCase):
    def setUp(self):
        self.agent_self_validation_call_name = (
            "runner.pcmk.resource_agent_self_validation"
        )
        self.env_assist, self.config = get_env_tools(test_case=self)
        self.config.runner.pcmk.load_agent()
        self.config.runner.cib.load()
        self.config.runner.pcmk.resource_agent_self_validation(
            {}, output="", name=self.agent_self_validation_call_name
        )

    def test_simplest_resource(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_simplest
            )
        )
        create_clone(self.env_assist.get_env(), wait=False)

    def test_custom_clone_id(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_custom_id
            )
        )
        create_clone(
            self.env_assist.get_env(), wait=False, clone_id="CustomCloneId"
        )

    def test_custom_clone_id_error_invalid_id(self):
        self.config.remove(self.agent_self_validation_call_name)
        self.env_assist.assert_raise_library_error(
            lambda: create_clone(
                self.env_assist.get_env(), wait=False, clone_id="1invalid"
            ),
        )
        self.env_assist.assert_reports(
            [fixture.report_invalid_id("1invalid", "1")],
        )

    def test_custom_clone_id_error_id_already_exist(self):
        self.config.remove(name="runner.cib.load")
        self.config.remove(self.agent_self_validation_call_name)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive class="ocf" id="C" provider="heartbeat"
                        type="Dummy"
                    >
                        <operations>
                            <op id="C-monitor-interval-10s" interval="10s"
                                name="monitor" timeout="20s"/>
                        </operations>
                    </primitive>
                </resources>
            """,
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_clone(
                self.env_assist.get_env(), wait=False, clone_id="C"
            ),
        )
        self.env_assist.assert_reports([fixture.report_id_already_exist("C")])

    def test_cib_upgrade_on_onfail_demote(self):
        self.config.remove(self.agent_self_validation_call_name)
        self.config.runner.cib.load(
            filename="cib-empty-3.3.xml",
            instead="runner.cib.load",
            name="load_cib_old_version",
        )
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(filename="cib-empty-3.4.xml")
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources="""<resources>
                <clone id="A-clone">
                    <primitive class="ocf" id="A" provider="heartbeat"
                        type="Dummy"
                    >
                        <operations>
                            <op id="A-migrate_from-interval-0s" interval="0s"
                                name="migrate_from" timeout="20"
                            />
                            <op id="A-migrate_to-interval-0s" interval="0s"
                                name="migrate_to" timeout="20"
                            />
                            <op id="A-monitor-interval-10" interval="10"
                                name="monitor" timeout="10" on-fail="demote"
                            />
                            <op id="A-reload-interval-0s" interval="0s"
                                name="reload" timeout="20"
                            />
                            <op id="A-start-interval-0s" interval="0s"
                                name="start" timeout="20"
                            />
                            <op id="A-stop-interval-0s" interval="0s"
                                name="stop" timeout="20"
                            />
                        </operations>
                    </primitive>
                </clone>
            </resources>"""
        )

        create_clone(
            self.env_assist.get_env(),
            operation_list=[
                {
                    "name": "monitor",
                    "timeout": "10",
                    "interval": "10",
                    "on-fail": "demote",
                }
            ],
            wait=False,
        )
        self.env_assist.assert_reports(
            [fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_fail_wait(self):
        self.config.env.push_cib(
            resources=fixture_cib_resources_xml_clone_simplest,
            wait=TIMEOUT,
            exception=LibraryError(
                reports.item.ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(wait_error_message)
                )
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_clone(self.env_assist.get_env()),
            [fixture.report_wait_for_idle_timed_out(wait_error_message)],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_fail(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_simplest, wait=TIMEOUT
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(failed="true")
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_clone(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_DOES_NOT_RUN, resource_id="A"
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_ok(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_simplest, wait=TIMEOUT
            ).runner.pcmk.load_state(resources=fixture_state_resources_xml())
        )
        create_clone(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_fail(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_simplest_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(resources=fixture_state_resources_xml())
        )

        self.env_assist.assert_raise_library_error(
            lambda: create_clone(self.env_assist.get_env(), disabled=True)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    roles_with_nodes={"Started": ["node1"]},
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok(self):
        (
            self.config.env.push_cib(
                resources=fixture_cib_resources_xml_clone_simplest_disabled,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_clone(self.env_assist.get_env(), disabled=True)
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok_by_target_role(self):
        (
            self.config.env.push_cib(
                resources="""
                    <resources>
                        <clone id="A-clone">
                            <primitive class="ocf" id="A" provider="heartbeat"
                                type="Dummy"
                            >
                                <meta_attributes id="A-meta_attributes">
                                    <nvpair id="A-meta_attributes-target-role"
                                        name="target-role"
                                        value="Stopped"
                                    />
                                </meta_attributes>
                                <operations>
                                    <op id="A-migrate_from-interval-0s"
                                        interval="0s" name="migrate_from"
                                        timeout="20"
                                    />
                                    <op id="A-migrate_to-interval-0s"
                                        interval="0s" name="migrate_to"
                                        timeout="20"
                                    />
                                    <op id="A-monitor-interval-10" interval="10"
                                        name="monitor" timeout="20"
                                    />
                                    <op id="A-reload-interval-0s" interval="0s"
                                        name="reload" timeout="20"
                                    />
                                    <op id="A-start-interval-0s" interval="0s"
                                        name="start" timeout="20"
                                    />
                                    <op id="A-stop-interval-0s" interval="0s"
                                        name="stop" timeout="20"
                                    />
                                </operations>
                            </primitive>
                        </clone>
                    </resources>
                """,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_clone(
            self.env_assist.get_env(),
            meta_attributes={"target-role": "Stopped"},
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok_by_target_role_in_clone(self):
        (
            self.config.env.push_cib(
                resources="""
                    <resources>
                        <clone id="A-clone">
                            <primitive class="ocf" id="A" provider="heartbeat"
                                type="Dummy"
                            >
                                <operations>
                                    <op id="A-migrate_from-interval-0s"
                                        interval="0s" name="migrate_from"
                                        timeout="20"
                                    />
                                    <op id="A-migrate_to-interval-0s"
                                        interval="0s" name="migrate_to"
                                        timeout="20"
                                    />
                                    <op id="A-monitor-interval-10" interval="10"
                                        name="monitor" timeout="20"
                                    />
                                    <op id="A-reload-interval-0s" interval="0s"
                                        name="reload" timeout="20"
                                    />
                                    <op id="A-start-interval-0s" interval="0s"
                                        name="start" timeout="20"
                                    />
                                    <op id="A-stop-interval-0s" interval="0s"
                                        name="stop" timeout="20"
                                    />
                                </operations>
                            </primitive>
                            <meta_attributes id="A-clone-meta_attributes">
                                <nvpair id="A-clone-meta_attributes-target-role"
                                    name="target-role" value="Stopped"
                                />
                            </meta_attributes>
                        </clone>
                    </resources>
                """,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_clone(
            self.env_assist.get_env(), clone_options={"target-role": "Stopped"}
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok_by_clone_max(self):
        (
            self.config.env.push_cib(
                resources="""
                    <resources>
                        <clone id="A-clone">
                            <primitive class="ocf" id="A" provider="heartbeat"
                                type="Dummy"
                            >
                                <operations>
                                    <op id="A-migrate_from-interval-0s"
                                        interval="0s" name="migrate_from"
                                        timeout="20"
                                    />
                                    <op id="A-migrate_to-interval-0s"
                                        interval="0s" name="migrate_to"
                                        timeout="20"
                                    />
                                    <op id="A-monitor-interval-10" interval="10"
                                        name="monitor" timeout="20"
                                    />
                                    <op id="A-reload-interval-0s" interval="0s"
                                        name="reload" timeout="20"
                                    />
                                    <op id="A-start-interval-0s" interval="0s"
                                        name="start" timeout="20"
                                    />
                                    <op id="A-stop-interval-0s" interval="0s"
                                        name="stop" timeout="20"
                                    />
                                </operations>
                            </primitive>
                            <meta_attributes id="A-clone-meta_attributes">
                                <nvpair id="A-clone-meta_attributes-clone-max"
                                    name="clone-max" value="0"
                                />
                            </meta_attributes>
                        </clone>
                    </resources>
                """,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_clone(
            self.env_assist.get_env(), clone_options={"clone-max": "0"}
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_disable_ok_by_clone_node_max(self):
        (
            self.config.env.push_cib(
                resources="""
                    <resources>
                        <clone id="A-clone">
                            <primitive class="ocf" id="A" provider="heartbeat"
                                type="Dummy"
                            >
                                <operations>
                                    <op id="A-migrate_from-interval-0s"
                                        interval="0s" name="migrate_from"
                                        timeout="20"
                                    />
                                    <op id="A-migrate_to-interval-0s"
                                        interval="0s" name="migrate_to"
                                        timeout="20"
                                    />
                                    <op id="A-monitor-interval-10" interval="10"
                                        name="monitor" timeout="20"
                                    />
                                    <op id="A-reload-interval-0s" interval="0s"
                                        name="reload" timeout="20"
                                    />
                                    <op id="A-start-interval-0s" interval="0s"
                                        name="start" timeout="20"
                                    />
                                    <op id="A-stop-interval-0s" interval="0s"
                                        name="stop" timeout="20"
                                    />
                                </operations>
                            </primitive>
                            <meta_attributes id="A-clone-meta_attributes">
                                <nvpair
                                    id="A-clone-meta_attributes-clone-node-max"
                                    name="clone-node-max" value="0"
                                />
                            </meta_attributes>
                        </clone>
                    </resources>
                """,
                wait=TIMEOUT,
            ).runner.pcmk.load_state(
                resources=fixture_state_resources_xml(role="Stopped")
            )
        )
        create_clone(
            self.env_assist.get_env(), clone_options={"clone-node-max": "0"}
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.RESOURCE_DOES_NOT_RUN,
                    resource_id="A",
                )
            ]
        )


class CreateInToBundle(TestCase):
    fixture_empty_resources = "<resources />"

    fixture_resources_pre = """
        <resources>
            <bundle id="B">
                <network control-port="12345" ip-range-start="192.168.100.200"/>
            </bundle>
        </resources>
    """

    fixture_resource_post_simple_without_network = """
        <resources>
            <bundle id="B">
                {network}
                <primitive
                    class="ocf" id="A" provider="heartbeat" type="Dummy"
                >
                    <operations>
                        <op id="A-migrate_from-interval-0s" interval="0s"
                            name="migrate_from" timeout="20"
                        />
                        <op id="A-migrate_to-interval-0s" interval="0s"
                            name="migrate_to" timeout="20"
                        />
                        <op id="A-monitor-interval-10" interval="10"
                            name="monitor" timeout="20" {onfail}
                        />
                        <op id="A-reload-interval-0s" interval="0s" name="reload"
                            timeout="20"
                        />
                        <op id="A-start-interval-0s" interval="0s"
                            name="start" timeout="20"
                        />
                        <op id="A-stop-interval-0s" interval="0s"
                            name="stop" timeout="20"
                        />
                    </operations>
                </primitive>
            </bundle>
        </resources>
    """

    # fmt: off
    fixture_resources_post_simple = (
        fixture_resource_post_simple_without_network.format(
            network="""
                <network control-port="12345" ip-range-start="192.168.100.200"/>
            """,
            onfail=""
        )
    )
    # fmt: on

    fixture_resources_post_disabled = """
        <resources>
            <bundle id="B">
                <network control-port="12345" ip-range-start="192.168.100.200"/>
                <primitive
                    class="ocf" id="A" provider="heartbeat" type="Dummy"
                >
                    <meta_attributes id="A-meta_attributes">
                        <nvpair id="A-meta_attributes-target-role"
                            name="target-role" value="Stopped"
                        />
                    </meta_attributes>
                    <operations>
                        <op id="A-migrate_from-interval-0s" interval="0s"
                            name="migrate_from" timeout="20"
                        />
                        <op id="A-migrate_to-interval-0s" interval="0s"
                            name="migrate_to" timeout="20"
                        />
                        <op id="A-monitor-interval-10" interval="10"
                            name="monitor" timeout="20"
                        />
                        <op id="A-reload-interval-0s" interval="0s" name="reload"
                            timeout="20"
                        />
                        <op id="A-start-interval-0s" interval="0s"
                            name="start" timeout="20"
                        />
                        <op id="A-stop-interval-0s" interval="0s"
                            name="stop" timeout="20"
                        />
                    </operations>
                </primitive>
            </bundle>
        </resources>
    """

    fixture_status_stopped = """
        <resources>
            <bundle id="B" managed="true">
                <replica id="0">
                    <resource id="B-0" managed="true" role="Stopped" />
                </replica>
            </bundle>
        </resources>
    """

    fixture_status_running_with_primitive = """
        <resources>
            <bundle id="B" managed="true">
                <replica id="0">
                    <resource id="B-0" managed="true" role="Started">
                        <node name="node1" id="1" cached="false"/>
                    </resource>
                    <resource id="A" managed="true" role="Started">
                        <node name="node1" id="1" cached="false"/>
                    </resource>
                </replica>
            </bundle>
        </resources>
    """

    fixture_status_primitive_not_running = """
        <resources>
            <bundle id="B" managed="true">
                <replica id="0">
                    <resource id="B-0" managed="true" role="Started">
                        <node name="node1" id="1" cached="false"/>
                    </resource>
                    <resource id="A" managed="true" role="Stopped"/>
                </replica>
            </bundle>
        </resources>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(
            test_case=self,
            base_cib_filename="cib-empty.xml",
        )
        self.config.runner.pcmk.load_agent()

    def test_cib_upgrade_on_onfail_demote(self):
        self.config.runner.cib.load(
            filename="cib-empty-3.3.xml",
            name="load_cib_old_version",
        )
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(
            filename="cib-empty-3.4.xml", resources=self.fixture_resources_pre
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resource_post_simple_without_network.format(
                network="""
                    <network
                        control-port="12345" ip-range-start="192.168.100.200"
                    />
                """,
                onfail='on-fail="demote"',
            )
        )

        create_bundle(
            self.env_assist.get_env(),
            operation_list=[
                {
                    "name": "monitor",
                    "timeout": "20",
                    "interval": "10",
                    "on-fail": "demote",
                }
            ],
            wait=False,
        )
        self.env_assist.assert_reports(
            [fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL)]
        )

    def test_simplest_resource(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(resources=self.fixture_resources_post_simple)
        create_bundle(self.env_assist.get_env(), wait=False)

    def test_bundle_doesnt_exist(self):
        self.config.runner.cib.load(resources=self.fixture_empty_resources)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env(), wait=False),
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="B",
                    expected_types=["bundle"],
                    context_type="resources",
                    context_id="",
                )
            ],
            expected_in_processor=False,
        )

    def test_id_not_bundle(self):
        self.config.runner.cib.load(
            resources="""
                    <resources>
                        <primitive id="B"/>
                    </resources>
                """
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")

        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env(), wait=False),
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="B",
                    expected_types=["bundle"],
                    current_type="primitive",
                )
            ],
            expected_in_processor=False,
        )

    def test_bundle_not_empty(self):
        self.config.runner.cib.load(
            resources="""
                    <resources>
                        <bundle id="B">
                            <network control-port="12345"/>
                            <primitive id="P"/>
                        </bundle>
                    </resources>
                """
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env(), wait=False),
            [
                fixture.error(
                    reports.codes.RESOURCE_BUNDLE_ALREADY_CONTAINS_A_RESOURCE,
                    bundle_id="B",
                    resource_id="P",
                )
            ],
            expected_in_processor=False,
        )

    def test_wait_fail(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resources_post_simple,
            wait=TIMEOUT,
            exception=LibraryError(
                reports.item.ReportItem.error(
                    reports.messages.WaitForIdleTimedOut(wait_error_message)
                )
            ),
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env()),
            [
                fixture.report_wait_for_idle_timed_out(wait_error_message),
            ],
            expected_in_processor=False,
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_ok(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resources_post_simple, wait=TIMEOUT
        )
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_running_with_primitive
        )
        create_bundle(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.report_resource_running("A", {"Started": ["node1"]}),
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_wait_ok_run_fail(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resources_post_simple, wait=TIMEOUT
        )
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_primitive_not_running
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_DOES_NOT_RUN, resource_id="A"
                )
            ]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_disabled_wait_ok_not_running(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resources_post_disabled, wait=TIMEOUT
        )
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_primitive_not_running
        )
        create_bundle(self.env_assist.get_env(), disabled=True)
        self.env_assist.assert_reports(
            [fixture.report_resource_not_running("A")]
        )

    @mock.patch.object(
        settings,
        "pacemaker_api_result_schema",
        rc("pcmk_api_rng/api-result.rng"),
    )
    def test_disabled_wait_ok_running(self):
        self.config.runner.cib.load(resources=self.fixture_resources_pre)
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=self.fixture_resources_post_disabled, wait=TIMEOUT
        )
        self.config.runner.pcmk.load_state(
            resources=self.fixture_status_running_with_primitive
        )
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env(), disabled=True)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_RUNNING_ON_NODES,
                    resource_id="A",
                    roles_with_nodes={"Started": ["node1"]},
                )
            ]
        )

    def test_no_port_no_ip(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <bundle id="B"/>
                </resources>
            """
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.env_assist.assert_raise_library_error(
            lambda: create_bundle(self.env_assist.get_env(), wait=False)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    # pylint: disable=line-too-long
                    reports.codes.RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE,
                    bundle_id="B",
                    inner_resource_id="A",
                    force_code=reports.codes.FORCE,
                )
            ]
        )

    def test_no_port_no_ip_forced(self):
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <bundle id="B"/>
                </resources>
            """
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=(
                self.fixture_resource_post_simple_without_network.format(
                    network="", onfail=""
                )
            )
        )
        create_bundle(
            self.env_assist.get_env(),
            wait=False,
            allow_not_accessible_resource=True,
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE,
                    bundle_id="B",
                    inner_resource_id="A",
                )
            ]
        )

    def _test_with_network_defined(self, network):
        self.config.runner.cib.load(
            resources=f"""
                <resources>
                    <bundle id="B">
                        {network}
                    </bundle>
                </resources>
            """
        )
        self.config.runner.pcmk.resource_agent_self_validation({}, output="")
        self.config.env.push_cib(
            resources=(
                self.fixture_resource_post_simple_without_network.format(
                    network=network, onfail=""
                )
            )
        )
        create_bundle(self.env_assist.get_env(), wait=False)

    def test_port_defined(self):
        self._test_with_network_defined('<network control-port="12345"/>')

    def test_ip_range_defined(self):
        self._test_with_network_defined(
            '<network ip-range-start="192.168.100.200"/>'
        )



from pcs.common import reports
from pcs.common.node_communicator import RequestData
from pcs.common.reports.item import ReportItem
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    RunRemotelyBase,
    SimpleResponseProcessingMixin,
    SkipOfflineMixin,
)


class QdeviceBase(
    SkipOfflineMixin, AllSameDataMixin, AllAtOnceStrategyMixin, RunRemotelyBase
):
    # pylint: disable=abstract-method
    def __init__(self, report_processor, skip_offline_targets=False):
        super().__init__(report_processor)
        self._set_skip_offline(skip_offline_targets)


class Stop(SimpleResponseProcessingMixin, QdeviceBase):
    # pylint: disable=too-many-ancestors
    def _get_request_data(self):
        return RequestData("remote/qdevice_client_stop")

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_STOP,
                "corosync-qdevice",
                node_label,
            )
        )


class Start(QdeviceBase):
    def _get_request_data(self):
        return RequestData("remote/qdevice_client_start")

    def _process_response(self, response):
        report = self._get_response_report(response)
        node_label = response.request.target.label
        if report is None:
            if response.data == "corosync is not running, skipping":
                report = ReportItem.info(
                    reports.messages.ServiceActionSkipped(
                        reports.const.SERVICE_ACTION_START,
                        "corosync-qdevice",
                        "corosync is not running",
                        node_label,
                    )
                )
            else:
                report = ReportItem.info(
                    reports.messages.ServiceActionSucceeded(
                        reports.const.SERVICE_ACTION_START,
                        "corosync-qdevice",
                        node_label,
                    )
                )
        self._report(report)


class Enable(QdeviceBase):
    def _get_request_data(self):
        return RequestData("remote/qdevice_client_enable")

    def _process_response(self, response):
        report = self._get_response_report(response)
        node_label = response.request.target.label
        if report is None:
            if response.data == "corosync is not enabled, skipping":
                report = ReportItem.info(
                    reports.messages.ServiceActionSkipped(
                        reports.const.SERVICE_ACTION_ENABLE,
                        "corosync-qdevice",
                        "corosync is not enabled",
                        node_label,
                    )
                )
            else:
                report = ReportItem.info(
                    reports.messages.ServiceActionSucceeded(
                        reports.const.SERVICE_ACTION_ENABLE,
                        "corosync-qdevice",
                        node_label,
                    )
                )
        self._report(report)


class Disable(SimpleResponseProcessingMixin, QdeviceBase):
    def _get_request_data(self):
        return RequestData("remote/qdevice_client_disable")

    def _get_success_report(self, node_label):
        return ReportItem.info(
            reports.messages.ServiceActionSucceeded(
                reports.const.SERVICE_ACTION_DISABLE,
                "corosync-qdevice",
                node_label,
            )
        )



from typing import (
    Any,
    Iterator,
    Optional,
    Tuple,
)

import pyparsing

from .expression_part import (
    BOOL_AND,
    BOOL_OR,
    DATE_OP_GT,
    DATE_OP_LT,
    NODE_ATTR_OP_DEFINED,
    NODE_ATTR_OP_EQ,
    NODE_ATTR_OP_GT,
    NODE_ATTR_OP_GTE,
    NODE_ATTR_OP_LT,
    NODE_ATTR_OP_LTE,
    NODE_ATTR_OP_NE,
    NODE_ATTR_OP_NOT_DEFINED,
    NODE_ATTR_TYPE_INTEGER,
    NODE_ATTR_TYPE_NUMBER,
    NODE_ATTR_TYPE_STRING,
    NODE_ATTR_TYPE_VERSION,
    BoolExpr,
    DateInRangeExpr,
    DatespecExpr,
    DateUnaryExpr,
    NodeAttrExpr,
    OpExpr,
    RscExpr,
    RuleExprPart,
)

_token_to_date_expr_unary_op = {
    "gt": DATE_OP_GT,
    "lt": DATE_OP_LT,
}

_token_to_node_expr_unary_op = {
    "defined": NODE_ATTR_OP_DEFINED,
    "not_defined": NODE_ATTR_OP_NOT_DEFINED,
}

_token_to_node_expr_binary_op = {
    "eq": NODE_ATTR_OP_EQ,
    "ne": NODE_ATTR_OP_NE,
    "gte": NODE_ATTR_OP_GTE,
    "gt": NODE_ATTR_OP_GT,
    "lte": NODE_ATTR_OP_LTE,
    "lt": NODE_ATTR_OP_LT,
}

_token_to_node_expr_type = {
    "integer": NODE_ATTR_TYPE_INTEGER,
    "number": NODE_ATTR_TYPE_NUMBER,
    "string": NODE_ATTR_TYPE_STRING,
    "version": NODE_ATTR_TYPE_VERSION,
}


pyparsing.ParserElement.enablePackrat()


class RuleParseError(Exception):
    def __init__(
        self,
        rule_string: str,
        rule_line: str,
        lineno: int,
        colno: int,
        pos: int,
        msg: str,
    ):
        super().__init__()
        self.rule_string = rule_string
        self.rule_line = rule_line
        self.lineno = lineno
        self.colno = colno
        self.pos = pos
        self.msg = msg


def parse_rule(rule_string: str) -> BoolExpr:
    """
    Parse a rule string and return a corresponding semantic tree

    rule_string -- the whole rule expression
    """
    if not rule_string:
        return BoolExpr(BOOL_AND, [])

    try:
        parsed = __get_rule_parser().parseString(rule_string, parseAll=True)[0]
    except pyparsing.ParseException as e:
        raise RuleParseError(
            rule_string,
            e.line,
            e.lineno,
            e.col,
            e.loc,
            e.args[2],
        ) from e

    if not isinstance(parsed, BoolExpr):
        # If we only got a representation on an inner rule element instead of a
        # rule element itself, wrap the result in a default AND-rule. (There is
        # only one expression so "and" vs. "or" doesn't really matter.)
        parsed = BoolExpr(BOOL_AND, [parsed])

    return parsed


def __operator_operands(
    token_list: pyparsing.ParseResults,
) -> Iterator[Tuple[Any, Any]]:
    # See pyparsing examples
    # https://github.com/pyparsing/pyparsing/blob/master/examples/eval_arith.py
    token_iterator = iter(token_list)
    while True:
        try:
            yield (next(token_iterator), next(token_iterator))
        except StopIteration:
            break


def __build_bool_tree(token_list: pyparsing.ParseResults) -> RuleExprPart:
    # See pyparsing examples
    # https://github.com/pyparsing/pyparsing/blob/master/examples/eval_arith.py
    token_to_operator = {
        "and": BOOL_AND,
        "or": BOOL_OR,
    }
    operand_left = token_list[0][0]
    last_operator: Optional[str] = None
    operand_list = []
    for operator, operand_right in __operator_operands(token_list[0][1:]):
        # In each iteration, we get a bool_op ("and" or "or") and the right
        # operand.
        if last_operator == operator or last_operator is None:
            # If we got the same operator as last time (or this is the first
            # one), stack all the operads so we can put them all into one
            # BoolExpr class.
            operand_list.append(operand_right)
        else:
            # The operator has changed. Put all the stacked operands into the
            # correct BoolExpr class and start the stacking again. The created
            # class is the left operand of the current operator.
            operand_left = BoolExpr(
                token_to_operator[last_operator], [operand_left] + operand_list
            )
            operand_list = [operand_right]
        last_operator = operator
    if operand_list and last_operator:
        # Use any of the remaining stacked operands.
        operand_left = BoolExpr(
            token_to_operator[last_operator], [operand_left] + operand_list
        )
    return operand_left


def __build_date_unary_expr(
    parse_result: pyparsing.ParseResults,
) -> RuleExprPart:
    # Those attrs are defined by setResultsName in date_unary_expr grammar rule
    return DateUnaryExpr(
        _token_to_date_expr_unary_op[parse_result.operator], parse_result.date
    )


def __build_date_inrange_expr(
    parse_result: pyparsing.ParseResults,
) -> RuleExprPart:
    # Those attrs are defined by setResultsName in date_inrange_expr grammar
    # rule
    return DateInRangeExpr(
        parse_result.date1 if parse_result.date1 else None,
        parse_result.date2 if parse_result.date2 else None,
        parse_result.duration.asList() if parse_result.duration else None,
    )


def __build_datespec_expr(
    parse_result: pyparsing.ParseResults,
) -> RuleExprPart:
    # Those attrs are defined by setResultsName in datespec_expr grammar rule
    return DatespecExpr(
        parse_result.datespec.asList() if parse_result.datespec else None
    )


def __build_node_attr_unary_expr(
    parse_result: pyparsing.ParseResults,
) -> RuleExprPart:
    # Those attrs are defined by setResultsName in node_attr_unary_expr grammar
    # rule
    return NodeAttrExpr(
        _token_to_node_expr_unary_op[parse_result.operator],
        parse_result.attr_name,
        None,
        None,
    )


def __build_node_attr_binary_expr(
    parse_result: pyparsing.ParseResults,
) -> RuleExprPart:
    # Those attrs are defined by setResultsName in node_attr_binary_expr
    # grammar rule
    return NodeAttrExpr(
        _token_to_node_expr_binary_op[parse_result.operator],
        parse_result.attr_name,
        parse_result.attr_value,
        _token_to_node_expr_type[parse_result.attr_type]
        if parse_result.attr_type
        else None,
    )


def __build_op_expr(parse_result: pyparsing.ParseResults) -> RuleExprPart:
    # Those attrs are defined by setResultsName in op_expr grammar rule
    return OpExpr(
        parse_result.name,
        # pyparsing-2.1.0 puts "interval_value" into parse_result.interval as
        # defined in the grammar AND it also puts "interval_value" into
        # parse_result. pyparsing-2.4.0 only puts "interval_value" into
        # parse_result. Not sure why, maybe it's a bug, maybe it's intentional.
        parse_result.interval_value if parse_result.interval_value else None,
    )


def __build_rsc_expr(parse_result: pyparsing.ParseResults) -> RuleExprPart:
    # Those attrs are defined by the regexp in rsc_expr grammar rule
    return RscExpr(
        parse_result.standard, parse_result.provider, parse_result.type
    )


def __get_date_common_parser_part() -> pyparsing.ParserElement:
    # This only checks for <name>=<value> and returns a list of 2-tuples. The
    # tuples are expected to be validated elsewhere.
    return pyparsing.OneOrMore(
        pyparsing.Group(
            pyparsing.And(
                [
                    # name
                    # It can by any string containing any characters except
                    # whitespace (token separator), '=' (name-value separator)
                    # and "()" (brackets).
                    pyparsing.Regex(r"[^=\s()]+").setName("<date part name>"),
                    # Suppress is needed so the '=' doesn't pollute the
                    # resulting structure produced automatically by pyparsing.
                    pyparsing.Suppress(
                        # no spaces allowed around the "="
                        pyparsing.Literal("=").leaveWhitespace()
                    ),
                    # value
                    # It can by any string containing any characters except
                    # whitespace (token separator) and "()" (brackets).
                    pyparsing.Regex(r"[^\s()]+").setName("<date part value>"),
                ]
            )
        )
    )


def __get_rule_parser() -> pyparsing.ParserElement:
    # This function defines the rule grammar

    # How to add new rule expressions:
    #   1 Create new grammar rules in a way similar to existing rsc_expr and
    #     op_expr. Use setName for better description of a grammar when printed.
    #     Use setResultsName for an easy access to parsed parts.
    #   2 Create new classes in expression_part module, probably one for each
    #     type of expression. Those are data containers holding the parsed data
    #     independent of the parser.
    #   3 Create builders for the new classes and connect them to created
    #     grammar rules using setParseAction.
    #   4 Add the new expressions into simple_expr definition.
    #   5 Test and debug the whole thing.

    node_attr_unary_expr = pyparsing.And(
        [
            # operator
            pyparsing.Or(
                [
                    pyparsing.CaselessKeyword(op).setName(f"'{op}'")
                    for op in _token_to_node_expr_unary_op
                ]
            ).setResultsName("operator"),
            # attribute name
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets).
            pyparsing.Regex(r"[^\s()]+")
            .setName("<attribute name>")
            .setResultsName("attr_name"),
        ]
    )
    node_attr_unary_expr.setParseAction(__build_node_attr_unary_expr)

    node_attr_binary_expr = pyparsing.And(
        [
            # attribute name
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets).
            pyparsing.Regex(r"[^\s()]+")
            .setName("<attribute name>")
            .setResultsName("attr_name"),
            # operator
            pyparsing.Or(
                [
                    pyparsing.CaselessKeyword(op).setName(f"'{op}'")
                    for op in _token_to_node_expr_binary_op
                ]
            ).setResultsName("operator"),
            # attribute type
            pyparsing.Optional(
                pyparsing.Or(
                    [
                        pyparsing.CaselessKeyword(type_).setName(f"'{type_}'")
                        for type_ in _token_to_node_expr_type
                    ]
                )
            )
            .setName("<attribute type>")
            .setResultsName("attr_type"),
            # attribute value
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets).
            pyparsing.Regex(r"[^\s()]+")
            .setName("<attribute value>")
            .setResultsName("attr_value"),
        ]
    )
    node_attr_binary_expr.setParseAction(__build_node_attr_binary_expr)

    date_unary_expr = pyparsing.And(
        [
            pyparsing.CaselessKeyword("date").setName("'date'"),
            # operator
            pyparsing.Or(
                [
                    pyparsing.CaselessKeyword(op).setName(f"'{op}'")
                    for op in _token_to_date_expr_unary_op
                ]
            ).setResultsName("operator"),
            # date
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets).
            # The actual value should be validated elsewhere.
            pyparsing.Regex(r"[^\s()]+")
            .setName("<date>")
            .setResultsName("date"),
        ]
    )
    date_unary_expr.setParseAction(__build_date_unary_expr)

    date_inrange_expr = pyparsing.And(
        [
            pyparsing.CaselessKeyword("date").setName("'date'"),
            pyparsing.CaselessKeyword("in_range").setName("'in_range'"),
            # date
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets).
            # The actual value should be validated elsewhere.
            # The Regex matches 'to'. In order to prevent that, FollowedBy is
            # used.
            pyparsing.Optional(
                pyparsing.And(
                    [
                        pyparsing.Regex(r"[^\s()]+")
                        .setName("[<date>]")
                        .setResultsName("date1"),
                        pyparsing.FollowedBy(
                            pyparsing.CaselessKeyword("to").setName("'to'")
                        ),
                    ]
                )
            ),
            pyparsing.CaselessKeyword("to").setName("'to'"),
            pyparsing.Or(
                [
                    # date
                    # It can by any string containing any characters except
                    # whitespace (token separator) and "()" (brackets).
                    # The actual value should be validated elsewhere.
                    pyparsing.Regex(r"[^\s()]+")
                    .setName("<date>")
                    .setResultsName("date2"),
                    # duration
                    pyparsing.And(
                        [
                            pyparsing.CaselessKeyword("duration").setName(
                                "'duration'"
                            ),
                            __get_date_common_parser_part().setResultsName(
                                "duration"
                            ),
                        ]
                    ),
                ]
            ),
        ]
    )
    date_inrange_expr.setParseAction(__build_date_inrange_expr)

    datespec_expr = pyparsing.And(
        [
            pyparsing.CaselessKeyword("date-spec").setName("'date-spec'"),
            __get_date_common_parser_part().setResultsName("datespec"),
        ]
    )
    datespec_expr.setParseAction(__build_datespec_expr)

    rsc_expr = pyparsing.And(
        [
            pyparsing.CaselessKeyword("resource").setName("'resource'"),
            # resource name
            # Up to three parts separated by ":". The parts can contain any
            # characters except whitespace (token separator), ":" (parts
            # separator) and "()" (brackets).
            pyparsing.Regex(
                r"(?P<standard>[^\s:()]+)?:(?P<provider>[^\s:()]+)?:(?P<type>[^\s:()]+)?"
            ).setName("<resource name>"),
        ]
    )
    rsc_expr.setParseAction(__build_rsc_expr)

    op_interval = pyparsing.And(
        [
            pyparsing.CaselessKeyword("interval").setName("'interval'"),
            # no spaces allowed around the "="
            pyparsing.Literal("=").leaveWhitespace(),
            # interval value: number followed by a time unit, no spaces allowed
            # between the number and the unit thanks to Combine being used
            pyparsing.Combine(
                pyparsing.And(
                    [
                        pyparsing.Word(pyparsing.nums),
                        pyparsing.Optional(pyparsing.Word(pyparsing.alphas)),
                    ]
                )
            )
            .setName("<integer>[<time unit>]")
            .setResultsName("interval_value"),
        ]
    )
    op_expr = pyparsing.And(
        [
            pyparsing.CaselessKeyword("op").setName("'op'"),
            # operation name
            # It can by any string containing any characters except whitespace
            # (token separator) and "()" (brackets). Operations are defined in
            # agents' metadata which we do not have access to (e.g. when the
            # user sets operation "my_check" and doesn't even specify agent's
            # name).
            pyparsing.Regex(r"[^\s()]+")
            .setName("<operation name>")
            .setResultsName("name"),
            pyparsing.Optional(op_interval).setResultsName("interval"),
        ]
    )
    op_expr.setParseAction(__build_op_expr)

    # Ordering matters here as the first expression which matches wins. This is
    # mostly not an issue as the expressions don't overlap and the grammar is
    # not ambiguous. There are, exceptions, however:
    # 1) date gt something
    #   This can be either a date_unary_expr or a node_attr_binary_expr. We
    #   want it to be a date expression. If the user wants it to be a node
    #   attribute expression, they can do it like this: 'date gt <type>
    #   something' where <type> is an item of _token_to_node_expr_type. That
    #   way, both date and node attribute expression can be realized.
    simple_expr = pyparsing.Or(
        [
            date_unary_expr,
            date_inrange_expr,
            datespec_expr,
            node_attr_unary_expr,
            node_attr_binary_expr,
            rsc_expr,
            op_expr,
        ]
    )

    # See pyparsing examples
    # https://github.com/pyparsing/pyparsing/blob/master/examples/simpleBool.py
    # https://github.com/pyparsing/pyparsing/blob/master/examples/eval_arith.py
    bool_operator = pyparsing.Or(
        [
            pyparsing.CaselessKeyword("and").setName("'and'"),
            pyparsing.CaselessKeyword("or").setName("'or'"),
        ]
    )
    bool_expr = pyparsing.infixNotation(
        simple_expr,
        # By putting both "and" and "or" in one tuple we say they have the same
        # priority. This is consistent with legacy pcs parsers. And it is how
        # it should be, they work as a glue between "simple_expr"s.
        [(bool_operator, 2, pyparsing.opAssoc.LEFT, __build_bool_tree)],
    )

    return pyparsing.Or([bool_expr, simple_expr])



from typing import (
    Any,
    List,
    Mapping,
    Optional,
)

from dataclasses import dataclass

from pcs.common.interface.dto import DataTransferObject
from pcs.common.reports.dto import ReportItemDto

from .types import CommunicationResultStatus as StatusType


@dataclass(frozen=True)
class InternalCommunicationResultDto(DataTransferObject):
    status: StatusType
    status_msg: Optional[str]
    report_list: List[ReportItemDto]
    data: Any


@dataclass(frozen=True)
class InternalCommunicationRequestOptionsDto(DataTransferObject):
    request_timeout: Optional[int]


@dataclass(frozen=True)
class InternalCommunicationRequestDto(DataTransferObject):
    options: InternalCommunicationRequestOptionsDto
    cmd: str
    cmd_data: Mapping[str, Any]



from typing import (
    Container,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.lib.cib import (
    resource,
    stonith,
)
from pcs.lib.cib.nvpair import (
    INSTANCE_ATTRIBUTES_TAG,
    get_value,
)
from pcs.lib.cib.resource.common import are_meta_disabled
from pcs.lib.cib.tools import (
    ElementNotFound,
    IdProvider,
    get_element_by_id,
)
from pcs.lib.commands.resource import (
    _ensure_disabled_after_wait,
    resource_environment,
)
from pcs.lib.communication.corosync import GetCorosyncOnlineTargets
from pcs.lib.communication.scsi import (
    Unfence,
    UnfenceMpath,
)
from pcs.lib.communication.tools import (
    AllSameDataMixin,
    run_and_raise,
)
from pcs.lib.env import (
    LibraryEnvironment,
    WaitType,
)
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pacemaker.live import (
    FenceHistoryCommandErrorException,
    fence_history_cleanup,
    fence_history_text,
    fence_history_update,
    is_fence_history_supported_management,
    is_getting_resource_digest_supported,
)
from pcs.lib.pacemaker.values import validate_id
from pcs.lib.resource_agent import (
    InvalidResourceAgentName,
    ResourceAgentError,
    ResourceAgentFacade,
    ResourceAgentFacadeFactory,
    ResourceAgentName,
    UnableToGetAgentMetadata,
    resource_agent_error_to_report_item,
)
from pcs.lib.validate import validate_add_remove_items
from pcs.lib.xml_tools import get_root


def _get_agent_facade(
    report_processor: reports.ReportProcessor,
    factory: ResourceAgentFacadeFactory,
    name: str,
    allow_absent_agent: bool,
) -> ResourceAgentFacade:
    try:
        if ":" in name:
            raise InvalidResourceAgentName(name)
        full_name = ResourceAgentName("stonith", None, name)
        return factory.facade_from_parsed_name(full_name)
    except UnableToGetAgentMetadata as e:
        if allow_absent_agent:
            report_processor.report(
                resource_agent_error_to_report_item(
                    e, reports.ReportItemSeverity.warning(), is_stonith=True
                )
            )
            return factory.void_facade_from_parsed_name(full_name)
        report_processor.report(
            resource_agent_error_to_report_item(
                e,
                reports.ReportItemSeverity.error(reports.codes.FORCE),
                is_stonith=True,
            )
        )
        raise LibraryError() from e
    except ResourceAgentError as e:
        report_processor.report(
            resource_agent_error_to_report_item(
                e, reports.ReportItemSeverity.error(), is_stonith=True
            )
        )
        raise LibraryError() from e


def create(
    env: LibraryEnvironment,
    stonith_id: str,
    stonith_agent_name: str,
    operations: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    wait: WaitType = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create stonith as resource in a cib.

    env -- provides all for communication with externals
    stonith_id -- an identifier of stonith resource
    stonith_agent_name -- contains name for the identification of agent
    operations -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for primitive/instance_attributes
    allow_absent_agent -- a flag for allowing agent not installed in a system
    allow_invalid_operation -- a flag for allowing to use operations that
        are not listed in a stonith agent metadata
    allow_invalid_instance_attributes -- a flag for allowing to use instance
        attributes that are not listed in a stonith agent metadata or for
        allowing to not use the instance_attributes that are required in
        stonith agent metadata
    use_default_operations -- a flag for stopping of adding default cib
        operations (specified in a stonith agent)
    ensure_disabled -- flag that keeps resource in target-role "Stopped"
    wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    runner = env.cmd_runner()
    agent_factory = ResourceAgentFacadeFactory(runner, env.report_processor)
    stonith_agent = _get_agent_facade(
        env.report_processor,
        agent_factory,
        stonith_agent_name,
        allow_absent_agent,
    )
    if stonith_agent.metadata.provides_unfencing:
        meta_attributes = dict(meta_attributes, provides="unfencing")

    with resource_environment(
        env,
        wait,
        [stonith_id],
        _ensure_disabled_after_wait(
            ensure_disabled or are_meta_disabled(meta_attributes),
        ),
    ) as resources_section:
        id_provider = IdProvider(resources_section)
        stonith_element = resource.primitive.create(
            env.report_processor,
            runner,
            resources_section,
            id_provider,
            stonith_id,
            stonith_agent,
            raw_operation_list=operations,
            meta_attributes=meta_attributes,
            instance_attributes=instance_attributes,
            allow_invalid_operation=allow_invalid_operation,
            allow_invalid_instance_attributes=allow_invalid_instance_attributes,
            use_default_operations=use_default_operations,
            resource_type="stonith",
        )
        if ensure_disabled:
            resource.common.disable(stonith_element, id_provider)


def create_in_group(
    env: LibraryEnvironment,
    stonith_id: str,
    stonith_agent_name: str,
    group_id: str,
    operations: Iterable[Mapping[str, str]],
    meta_attributes: Mapping[str, str],
    instance_attributes: Mapping[str, str],
    allow_absent_agent: bool = False,
    allow_invalid_operation: bool = False,
    allow_invalid_instance_attributes: bool = False,
    use_default_operations: bool = True,
    ensure_disabled: bool = False,
    adjacent_resource_id: Optional[str] = None,
    put_after_adjacent: bool = False,
    wait: WaitType = False,
):
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Create stonith as resource in a cib and put it into defined group.

    env -- provides all for communication with externals
    stonith_id --an identifier of stonith resource
    stonith_agent_name -- contains name for the identification of agent
    group_id -- identificator for group to put stonith inside
    operations -- contains attributes for each entered operation
    meta_attributes -- contains attributes for primitive/meta_attributes
    instance_attributes -- contains attributes for primitive/instance_attributes
    allow_absent_agent -- a flag for allowing agent not installed in a system
    allow_invalid_operation -- a flag for allowing to use operations that
        are not listed in a stonith agent metadata
    allow_invalid_instance_attributes -- a flag for allowing to use instance
        attributes that are not listed in a stonith agent metadata or for
        allowing to not use the instance_attributes that are required in
        stonith agent metadata
    use_default_operations -- a flag for stopping of adding default cib
        operations (specified in a stonith agent)
    ensure_disabled -- flag that keeps resource in target-role "Stopped"
    adjacent_resource_id -- identify neighbor of a newly created stonith
    put_after_adjacent -- is flag to put a newly create resource befor/after
        adjacent stonith
    wait -- flag for controlling waiting for pacemaker idle mechanism
    """
    runner = env.cmd_runner()
    agent_factory = ResourceAgentFacadeFactory(runner, env.report_processor)
    stonith_agent = _get_agent_facade(
        env.report_processor,
        agent_factory,
        stonith_agent_name,
        allow_absent_agent,
    )
    if stonith_agent.metadata.provides_unfencing:
        meta_attributes = dict(meta_attributes, provides="unfencing")

    with resource_environment(
        env,
        wait,
        [stonith_id],
        _ensure_disabled_after_wait(
            ensure_disabled or are_meta_disabled(meta_attributes),
        ),
    ) as resources_section:
        id_provider = IdProvider(resources_section)

        adjacent_resource_element = None
        if adjacent_resource_id:
            try:
                adjacent_resource_element = get_element_by_id(
                    get_root(resources_section), adjacent_resource_id
                )
            except ElementNotFound:
                # We cannot continue without adjacent element because
                # the validator might produce misleading reports
                if env.report_processor.report(
                    ReportItem.error(
                        reports.messages.IdNotFound(adjacent_resource_id, [])
                    )
                ).has_errors:
                    raise LibraryError() from None

        try:
            group_element = get_element_by_id(
                get_root(resources_section), group_id
            )
        except ElementNotFound:
            group_id_reports: List[ReportItem] = []
            validate_id(
                group_id, description="group name", reporter=group_id_reports
            )
            env.report_processor.report_list(group_id_reports)
            group_element = resource.group.append_new(
                resources_section, group_id
            )

        stonith_element = resource.primitive.create(
            env.report_processor,
            runner,
            resources_section,
            id_provider,
            stonith_id,
            stonith_agent,
            operations,
            meta_attributes,
            instance_attributes,
            allow_invalid_operation,
            allow_invalid_instance_attributes,
            use_default_operations,
        )
        if ensure_disabled:
            resource.common.disable(stonith_element, id_provider)

        if env.report_processor.report_list(
            resource.hierarchy.validate_move_resources_to_group(
                group_element,
                [stonith_element],
                adjacent_resource_element,
            )
        ).has_errors:
            raise LibraryError()

        resource.hierarchy.move_resources_to_group(
            group_element,
            [stonith_element],
            adjacent_resource_element,
            put_after_adjacent,
        )


def history_get_text(env: LibraryEnvironment, node: Optional[str] = None):
    """
    Get full fencing history in plain text

    env
    node -- get history for the specified node or all nodes if None
    """
    runner = env.cmd_runner()
    if not is_fence_history_supported_management(runner):
        raise LibraryError(
            ReportItem.error(reports.messages.FenceHistoryNotSupported())
        )

    try:
        return fence_history_text(runner, node)
    except FenceHistoryCommandErrorException as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.FenceHistoryCommandError(
                    str(e), reports.const.FENCE_HISTORY_COMMAND_SHOW
                )
            )
        ) from e


def history_cleanup(env: LibraryEnvironment, node: Optional[str] = None):
    """
    Clear fencing history

    env
    node -- clear history for the specified node or all nodes if None
    """
    runner = env.cmd_runner()
    if not is_fence_history_supported_management(runner):
        raise LibraryError(
            ReportItem.error(reports.messages.FenceHistoryNotSupported())
        )

    try:
        return fence_history_cleanup(runner, node)
    except FenceHistoryCommandErrorException as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.FenceHistoryCommandError(
                    str(e), reports.const.FENCE_HISTORY_COMMAND_CLEANUP
                )
            )
        ) from e


def history_update(env: LibraryEnvironment):
    """
    Update fencing history in a cluster (sync with other nodes)

    env
    """
    runner = env.cmd_runner()
    if not is_fence_history_supported_management(runner):
        raise LibraryError(
            ReportItem.error(reports.messages.FenceHistoryNotSupported())
        )

    try:
        return fence_history_update(runner)
    except FenceHistoryCommandErrorException as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.FenceHistoryCommandError(
                    str(e), reports.const.FENCE_HISTORY_COMMAND_UPDATE
                )
            )
        ) from e


def _update_scsi_devices_get_element_and_devices(
    runner: CommandRunner,
    report_processor: ReportProcessor,
    cib: _Element,
    stonith_id: str,
) -> Tuple[_Element, List[str]]:
    """
    Do checks and return stonith element and list of current scsi devices.
    Raise LibraryError if checks fail.

    runner -- command runner instance
    report_processor -- tool for warning/info/error reporting
    cib -- cib element
    stonith_id -- id of stonith resource
    """
    if not is_getting_resource_digest_supported(runner):
        raise LibraryError(
            ReportItem.error(
                reports.messages.StonithRestartlessUpdateOfScsiDevicesNotSupported()
            )
        )
    (
        stonith_el,
        report_list,
    ) = stonith.validate_stonith_restartless_update(cib, stonith_id)
    if report_processor.report_list(report_list).has_errors:
        raise LibraryError()
    # for mypy, this should not happen because exception would be raised
    if stonith_el is None:
        raise AssertionError("stonith element is None")
    current_device_list = get_value(
        INSTANCE_ATTRIBUTES_TAG, stonith_el, "devices"
    )
    if current_device_list is None:
        raise AssertionError("current_device_list is None")
    return stonith_el, current_device_list.split(",")


def _unfencing_scsi_devices(
    env: LibraryEnvironment,
    stonith_el: _Element,
    original_devices: Iterable[str],
    updated_devices: Iterable[str],
    force_flags: Container[reports.types.ForceCode] = (),
) -> None:
    """
    Unfence scsi devices provided in device_list if it is possible to connect
    to pcsd and corosync is running.

    env -- provides all for communication with externals
    original_devices -- devices before update
    updated_devices -- devices after update
    force_flags -- list of flags codes
    """
    devices_to_unfence = set(updated_devices) - set(original_devices)
    if not devices_to_unfence:
        return
    cluster_nodes_names, nodes_report_list = get_existing_nodes_names(
        env.get_corosync_conf(),
        error_on_missing_name=True,
    )
    env.report_processor.report_list(nodes_report_list)
    (
        target_report_list,
        cluster_nodes_target_list,
    ) = env.get_node_target_factory().get_target_list_with_reports(
        cluster_nodes_names,
        allow_skip=False,
    )
    env.report_processor.report_list(target_report_list)
    if env.report_processor.has_errors:
        raise LibraryError()
    com_cmd: AllSameDataMixin = GetCorosyncOnlineTargets(
        env.report_processor,
        skip_offline_targets=reports.codes.SKIP_OFFLINE_NODES in force_flags,
    )
    com_cmd.set_targets(cluster_nodes_target_list)
    online_corosync_target_list = run_and_raise(
        env.get_node_communicator(), com_cmd
    )
    if stonith_el.get("type") == "fence_mpath":
        com_cmd = UnfenceMpath(
            env.report_processor,
            original_devices=sorted(original_devices),
            updated_devices=sorted(updated_devices),
            node_key_map=stonith.get_node_key_map_for_mpath(
                stonith_el,
                [target.label for target in online_corosync_target_list],
            ),
        )
    else:  # fence_scsi
        com_cmd = Unfence(
            env.report_processor,
            original_devices=sorted(original_devices),
            updated_devices=sorted(updated_devices),
        )
    com_cmd.set_targets(online_corosync_target_list)
    run_and_raise(env.get_node_communicator(), com_cmd)


def update_scsi_devices(
    env: LibraryEnvironment,
    stonith_id: str,
    set_device_list: Iterable[str],
    force_flags: Container[reports.types.ForceCode] = (),
) -> None:
    """
    Update scsi fencing devices without restart and affecting other resources.

    env -- provides all for communication with externals
    stonith_id -- id of stonith resource
    set_device_list -- paths to the scsi devices that would be set for stonith
        resource
    force_flags -- list of flags codes
    """
    if not set_device_list:
        env.report_processor.report(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "devices", "", None, cannot_be_empty=True
                )
            )
        )
    runner = env.cmd_runner()
    (
        stonith_el,
        current_device_list,
    ) = _update_scsi_devices_get_element_and_devices(
        runner, env.report_processor, env.get_cib(), stonith_id
    )
    if env.report_processor.has_errors:
        raise LibraryError()
    stonith.update_scsi_devices_without_restart(
        runner,
        env.get_cluster_state(),
        stonith_el,
        IdProvider(stonith_el),
        set_device_list,
    )
    _unfencing_scsi_devices(
        env, stonith_el, current_device_list, set_device_list, force_flags
    )
    env.push_cib()


def update_scsi_devices_add_remove(
    env: LibraryEnvironment,
    stonith_id: str,
    add_device_list: Iterable[str],
    remove_device_list: Iterable[str],
    force_flags: Container[reports.types.ForceCode] = (),
) -> None:
    """
    Update scsi fencing devices without restart and affecting other resources.

    env -- provides all for communication with externals
    stonith_id -- id of stonith resource
    add_device_list -- paths to the scsi devices that would be added to the
        stonith resource
    remove_device_list -- paths to the scsi devices that would be removed from
        the stonith resource
    force_flags -- list of flags codes
    """
    runner = env.cmd_runner()
    (
        stonith_el,
        current_device_list,
    ) = _update_scsi_devices_get_element_and_devices(
        runner, env.report_processor, env.get_cib(), stonith_id
    )
    if env.report_processor.report_list(
        validate_add_remove_items(
            add_device_list,
            remove_device_list,
            current_device_list,
            reports.const.ADD_REMOVE_CONTAINER_TYPE_STONITH_RESOURCE,
            reports.const.ADD_REMOVE_ITEM_TYPE_DEVICE,
            stonith_el.get("id", ""),
        )
    ).has_errors:
        raise LibraryError()
    updated_device_set = (
        set(current_device_list)
        .union(add_device_list)
        .difference(remove_device_list)
    )
    stonith.update_scsi_devices_without_restart(
        env.cmd_runner(),
        env.get_cluster_state(),
        stonith_el,
        IdProvider(stonith_el),
        updated_device_set,
    )
    _unfencing_scsi_devices(
        env, stonith_el, current_device_list, updated_device_set, force_flags
    )
    env.push_cib()



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.lib import resource_agent as ra
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent.types import (
    ResourceAgentActionOcf1_0,
    ResourceAgentActionOcf1_1,
    ResourceAgentMetadataOcf1_0,
    ResourceAgentMetadataOcf1_1,
    ResourceAgentParameterOcf1_0,
    ResourceAgentParameterOcf1_1,
)

from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.xml import etree_to_str


class LoadMetadataXml(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        metadata = """
            <resource-agent name="Dummy">
            </resource-agent>
        """
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            stdout=metadata,
        )

        env = self.env_assist.get_env()
        self.assertEqual(
            # pylint: disable=protected-access
            ra.xml._load_metadata_xml(env.cmd_runner(), agent_name),
            metadata.strip(),
        )

    def test_failure(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            agent_is_missing=True,
            stderr="error message",
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            # pylint: disable=protected-access
            ra.xml._load_metadata_xml(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "ocf:pacemaker:Dummy")
        self.assertEqual(cm.exception.message, "error message")


class LoadFakeAgentMetadataXml(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        agent_name = ra.const.PACEMAKER_FENCED
        metadata = """
            <resource-agent name="pacemaker-fenced">
            </resource-agent>
        """
        self.config.runner.pcmk.load_fenced_metadata(stdout=metadata)

        env = self.env_assist.get_env()
        self.assertEqual(
            # pylint: disable=protected-access
            ra.xml._load_fake_agent_metadata_xml(env.cmd_runner(), agent_name),
            metadata.strip(),
        )

    def test_failure(self):
        agent_name = ra.const.PACEMAKER_FENCED
        self.config.runner.pcmk.load_fenced_metadata(
            stdout="", stderr="error message"
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            # pylint: disable=protected-access
            ra.xml._load_fake_agent_metadata_xml(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "pacemaker-fenced")
        self.assertEqual(cm.exception.message, "error message")

    def test_unknown_agent(self):
        agent_name = "unknown"

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            # pylint: disable=protected-access
            ra.xml._load_fake_agent_metadata_xml(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "unknown")
        self.assertEqual(cm.exception.message, "Unknown agent")


class GetOcfVersion(TestCase):
    # pylint: disable=protected-access
    def test_no_version_element(self):
        self.assertEqual(
            ra.xml._get_ocf_version(
                etree.fromstring(
                    """
                        <resource-agent>
                        </resource-agent>
                    """
                )
            ),
            ra.const.OCF_1_0,
        )

    def test_version_element_empty(self):
        self.assertEqual(
            ra.xml._get_ocf_version(
                etree.fromstring(
                    """
                        <resource-agent>
                            <version/>
                        </resource-agent>
                    """
                )
            ),
            "",
        )

    def test_version_set(self):
        self.assertEqual(
            ra.xml._get_ocf_version(
                etree.fromstring(
                    """
                        <resource-agent>
                            <version> my version </version>
                        </resource-agent>
                    """
                )
            ),
            "my version",
        )

    def test_ignore_agent_version(self):
        self.assertEqual(
            ra.xml._get_ocf_version(
                etree.fromstring(
                    """
                        <resource-agent version="1.0">
                            <version>2.0</version>
                        </resource-agent>
                    """
                )
            ),
            "2.0",
        )


class MetadataXmlToDom(TestCase):
    # pylint: disable=protected-access
    def test_not_xml(self):
        with self.assertRaises(etree.XMLSyntaxError):
            ra.xml._metadata_xml_to_dom("not an xml")

    def test_no_version_not_valid(self):
        # pylint: disable=no-self-use
        metadata = """
            <resource-agent/>
        """
        assert_xml_equal(
            metadata, etree_to_str(ra.xml._metadata_xml_to_dom(metadata))
        )

    def test_no_version_valid(self):
        # pylint: disable=no-self-use
        metadata = """
            <resource-agent name="agent">
            </resource-agent>
        """
        assert_xml_equal(
            metadata, etree_to_str(ra.xml._metadata_xml_to_dom(metadata))
        )

    def test_ocf_1_0_not_valid(self):
        # pylint: disable=no-self-use
        metadata = """
            <resource-agent>
                <version>1.0</version>
            </resource-agent>
        """
        assert_xml_equal(
            metadata, etree_to_str(ra.xml._metadata_xml_to_dom(metadata))
        )

    def test_ocf_1_0_valid(self):
        # pylint: disable=no-self-use
        metadata = """
            <resource-agent name="agent">
                <version>1.0</version>
            </resource-agent>
        """
        assert_xml_equal(
            metadata, etree_to_str(ra.xml._metadata_xml_to_dom(metadata))
        )

    def test_ocf_1_1_not_valid(self):
        with self.assertRaises(etree.DocumentInvalid):
            ra.xml._metadata_xml_to_dom(
                """
                    <resource-agent>
                        <version>1.1</version>
                    </resource-agent>
                """
            )

    def test_ocf_1_1_valid(self):
        # pylint: disable=no-self-use
        metadata = """
            <resource-agent name="agent">
                <version>1.1</version>
                <parameters>
                    <parameter name="test" unique-group="ug1"/>
                </parameters>
            </resource-agent>
        """
        assert_xml_equal(
            metadata, etree_to_str(ra.xml._metadata_xml_to_dom(metadata))
        )


class LoadMetadata(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        metadata = """
            <resource-agent name="Dummy">
            </resource-agent>
        """
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            stdout=metadata,
        )

        env = self.env_assist.get_env()
        assert_xml_equal(
            metadata,
            etree_to_str(ra.xml.load_metadata(env.cmd_runner(), agent_name)),
        )

    def test_cannot_load(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            agent_is_missing=True,
            stderr="error message",
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.xml.load_metadata(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "ocf:pacemaker:Dummy")
        self.assertEqual(cm.exception.message, "error message")

    def test_not_xml(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            stdout="this is not an xml",
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.xml.load_metadata(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "ocf:pacemaker:Dummy")
        self.assertTrue(cm.exception.message.startswith("Start tag expected"))

    def test_not_valid_xml(self):
        agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
        metadata = "<resource-agent/>"
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:pacemaker:Dummy",
            stdout=metadata,
        )

        env = self.env_assist.get_env()
        assert_xml_equal(
            metadata,
            etree_to_str(ra.xml.load_metadata(env.cmd_runner(), agent_name)),
        )


class LoadFakeAgentMetadata(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        agent_name = ra.const.PACEMAKER_FENCED
        metadata = """
            <resource-agent name="pacemaker-fenced">
            </resource-agent>
        """
        self.config.runner.pcmk.load_fenced_metadata(stdout=metadata)

        env = self.env_assist.get_env()
        assert_xml_equal(
            metadata,
            etree_to_str(
                ra.xml.load_fake_agent_metadata(env.cmd_runner(), agent_name)
            ),
        )

    def test_cannot_load(self):
        agent_name = ra.const.PACEMAKER_FENCED
        self.config.runner.pcmk.load_fenced_metadata(
            stdout="", stderr="error message"
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.xml.load_fake_agent_metadata(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "pacemaker-fenced")
        self.assertEqual(cm.exception.message, "error message")

    def test_not_xml(self):
        agent_name = ra.const.PACEMAKER_FENCED
        self.config.runner.pcmk.load_fenced_metadata(
            stdout="this is not an xml"
        )

        env = self.env_assist.get_env()
        with self.assertRaises(ra.UnableToGetAgentMetadata) as cm:
            ra.xml.load_fake_agent_metadata(env.cmd_runner(), agent_name)
        self.assertEqual(cm.exception.agent_name, "pacemaker-fenced")
        self.assertTrue(cm.exception.message.startswith("Start tag expected"))

    def test_not_valid_xml(self):
        agent_name = ra.const.PACEMAKER_FENCED
        metadata = "<resource-agent/>"
        self.config.runner.pcmk.load_fenced_metadata(stdout=metadata)

        env = self.env_assist.get_env()
        assert_xml_equal(
            metadata,
            etree_to_str(
                ra.xml.load_fake_agent_metadata(env.cmd_runner(), agent_name)
            ),
        )


class ParseOcfToolsMixin:
    agent_name = ra.ResourceAgentName("ocf", "pacemaker", "Dummy")
    ocf_version = None
    parsed_ocf_version = None

    def parse(self, xml, agent_name=None):
        agent_name = agent_name or self.agent_name
        if not agent_name:
            raise AssertionError(
                "Invalid test usage, agent_name must be specified"
            )
        with mock.patch(
            "pcs.lib.resource_agent.xml._load_metadata_xml"
        ) as mock_load:
            mock_load.return_value = xml
            return ra.xml.parse_metadata(
                agent_name,
                ra.xml.load_metadata(
                    mock.MagicMock(spec=CommandRunner), agent_name
                ),
            )

    def xml(self, xml, agent_name=None, ocf_version=None):
        agent_name = agent_name or self.agent_name
        ocf_version = ocf_version or self.ocf_version
        dom = etree.fromstring(xml)
        if agent_name:
            dom.set("name", agent_name.full_name)
        if ocf_version:
            version_el = dom.find("./version")
            if version_el is None:
                version_el = etree.Element("version")
                dom.insert(0, version_el)
            version_el.text = ocf_version
        return etree_to_str(dom)

    def assert_parse_result(self, xml, metadata):
        self.assertEqual(
            self.parse(xml),
            (metadata, self.parsed_ocf_version or self.ocf_version),
        )


class ParseOcf10BaseMixin(ParseOcfToolsMixin):
    def test_empty_agent(self):
        self.assert_parse_result(
            self.xml("""<resource-agent/>"""),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <shortdesc>This is a shortdesc</shortdesc>
                            <longdesc>This is a longdesc</longdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="This is a shortdesc",
                longdesc="This is a longdesc",
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element_empty(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <longdesc/>
                            <shortdesc/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_attribute(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent shortdesc="This is a shortdesc">
                            <longdesc></longdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="This is a shortdesc",
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_attribute_empty(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent shortdesc=""/>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="",
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element_and_attribute(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent shortdesc="shortdesc attribute">
                            <shortdesc>shortdesc element</shortdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="shortdesc element",
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element_empty_and_attribute(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent shortdesc="shortdesc attribute">
                            <shortdesc></shortdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="shortdesc attribute",
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element_empty_and_attribute_empty(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent shortdesc="">
                            <shortdesc></shortdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc="",
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_parameters_empty_list(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_parameters_empty_parameter(self):
        self.assert_parse_result(
            self.xml(
                """
                    <resource-agent>
                        <parameters>
                            <parameter/>
                        </parameters>
                    </resource-agent>
                """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_0(
                        name="",
                        shortdesc=None,
                        longdesc=None,
                        type="string",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=None,
                        obsoletes=None,
                        unique=None,
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_minimal(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter"/>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_0(
                        name="a_parameter",
                        shortdesc=None,
                        longdesc=None,
                        type="string",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=None,
                        obsoletes=None,
                        unique=None,
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_all_settings(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter" required="1"
                                    unique="0" deprecated="1" obsoletes="old"
                                >
                                    <longdesc>Long description</longdesc>
                                    <shortdesc>short description</shortdesc>
                                    <content type="integer" default="123"/>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_0(
                        name="a_parameter",
                        shortdesc="short description",
                        longdesc="Long description",
                        type="integer",
                        default="123",
                        enum_values=None,
                        required="1",
                        deprecated="1",
                        obsoletes="old",
                        unique="0",
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_content(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="with_type">
                                    <content type="integer"/>
                                </parameter>
                                <parameter name="with_select">
                                    <content type="select" default="b">
                                        <option value="a"/>
                                        <option value="b"/>
                                        <option value="c"/>
                                    </content>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_0(
                        name="with_type",
                        shortdesc=None,
                        longdesc=None,
                        type="integer",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=None,
                        obsoletes=None,
                        unique=None,
                    ),
                    ResourceAgentParameterOcf1_0(
                        name="with_select",
                        shortdesc=None,
                        longdesc=None,
                        type="select",
                        default="b",
                        enum_values=["a", "b", "c"],
                        required=None,
                        deprecated=None,
                        obsoletes=None,
                        unique=None,
                    ),
                ],
                actions=[],
            ),
        )

    def test_actions_empty_list(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <actions/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_actions_empty_action(self):
        self.assert_parse_result(
            self.xml(
                """
                    <resource-agent>
                        <actions>
                            <action/>
                        </actions>
                    </resource-agent>
                """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[
                    ResourceAgentActionOcf1_0(
                        name="",
                        timeout=None,
                        interval=None,
                        role=None,
                        start_delay=None,
                        depth=None,
                        automatic=None,
                        on_target=None,
                    ),
                ],
            ),
        )

    def test_actions_multiple(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <actions>
                                <action name="minimal"/>
                                <action name="maximal" timeout="1" interval="2"
                                    start-delay="3" depth="4" role="Master"
                                />
                                <action name="stonith_special"
                                    automatic="0" on_target="1"
                                />
                            </actions>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_0(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[
                    ResourceAgentActionOcf1_0(
                        name="minimal",
                        timeout=None,
                        interval=None,
                        role=None,
                        start_delay=None,
                        depth=None,
                        automatic=None,
                        on_target=None,
                    ),
                    ResourceAgentActionOcf1_0(
                        name="maximal",
                        timeout="1",
                        interval="2",
                        role="Master",
                        start_delay="3",
                        depth="4",
                        automatic=None,
                        on_target=None,
                    ),
                    ResourceAgentActionOcf1_0(
                        name="stonith_special",
                        timeout=None,
                        interval=None,
                        role=None,
                        start_delay=None,
                        depth=None,
                        automatic="0",
                        on_target="1",
                    ),
                ],
            ),
        )


class ParseOcf10NoVersion(ParseOcf10BaseMixin, TestCase):
    parsed_ocf_version = "1.0"


class ParseOcf10UnsupportedVersion(ParseOcf10BaseMixin, TestCase):
    ocf_version = "0.1.2"


class ParseOcf10ExplicitVersion(ParseOcf10BaseMixin, TestCase):
    ocf_version = "1.0"


class ParseOcf11(ParseOcfToolsMixin, TestCase):
    ocf_version = "1.1"

    def test_empty_agent(self):
        self.assert_parse_result(
            self.xml("""<resource-agent/>"""),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <shortdesc>This is a shortdesc</shortdesc>
                            <longdesc>This is a longdesc</longdesc>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc="This is a shortdesc",
                longdesc="This is a longdesc",
                parameters=[],
                actions=[],
            ),
        )

    def test_desc_element_empty(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <longdesc/>
                            <shortdesc/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_parameters_empty_list(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_parameters_empty_parameter(self):
        # parameters must have at least 'name' attribute
        with self.assertRaises(ra.UnableToGetAgentMetadata):
            self.parse(
                self.xml(
                    """
                        <resource-agent>
                            <parameters>
                                <parameter/>
                            </parameters>
                        </resource-agent>
                    """
                )
            )

    def test_parameters_minimal(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter"/>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_1(
                        name="a_parameter",
                        shortdesc=None,
                        longdesc=None,
                        type="string",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=None,
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_deprecated_minimal(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter">
                                    <deprecated/>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_1(
                        name="a_parameter",
                        shortdesc=None,
                        longdesc=None,
                        type="string",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=True,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=None,
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_deprecated_replaced_with(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter">
                                    <deprecated>
                                        <replaced-with name="new1"/>
                                        <replaced-with name="new2"/>
                                    </deprecated>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_1(
                        name="a_parameter",
                        shortdesc=None,
                        longdesc=None,
                        type="string",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=True,
                        deprecated_by=["new1", "new2"],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=None,
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_all_settings(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="a_parameter"
                                    unique-group="ug1" unique="0" required="1"
                                    reloadable="0"
                                >
                                    <longdesc>Long description</longdesc>
                                    <shortdesc>short description</shortdesc>
                                    <deprecated>
                                        <replaced-with name="new1"/>
                                        <replaced-with name="new2"/>
                                        <desc>deprecation explanation</desc>
                                    </deprecated>
                                    <content type="integer" default="123"/>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_1(
                        name="a_parameter",
                        shortdesc="short description",
                        longdesc="Long description",
                        type="integer",
                        default="123",
                        enum_values=None,
                        required="1",
                        deprecated=True,
                        deprecated_by=["new1", "new2"],
                        deprecated_desc="deprecation explanation",
                        unique_group="ug1",
                        reloadable="0",
                    )
                ],
                actions=[],
            ),
        )

    def test_parameters_content(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <parameters>
                                <parameter name="with_type">
                                    <content type="integer"/>
                                </parameter>
                                <parameter name="with_select">
                                    <content type="select" default="b">
                                        <option value="a"/>
                                        <option value="b"/>
                                        <option value="c"/>
                                    </content>
                                </parameter>
                            </parameters>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[
                    ResourceAgentParameterOcf1_1(
                        name="with_type",
                        shortdesc=None,
                        longdesc=None,
                        type="integer",
                        default=None,
                        enum_values=None,
                        required=None,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=None,
                    ),
                    ResourceAgentParameterOcf1_1(
                        name="with_select",
                        shortdesc=None,
                        longdesc=None,
                        type="select",
                        default="b",
                        enum_values=["a", "b", "c"],
                        required=None,
                        deprecated=False,
                        deprecated_by=[],
                        deprecated_desc=None,
                        unique_group=None,
                        reloadable=None,
                    ),
                ],
                actions=[],
            ),
        )

    def test_actions_empty_list(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <actions/>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[],
            ),
        )

    def test_actions_empty_action(self):
        # actions must have at least 'name' attribute
        with self.assertRaises(ra.UnableToGetAgentMetadata):
            self.parse(
                self.xml(
                    """
                        <resource-agent>
                            <actions>
                                <action/>
                            </actions>
                        </resource-agent>
                    """
                )
            )

    def test_actions_multiple(self):
        self.assert_parse_result(
            self.xml(
                """
                        <resource-agent>
                            <actions>
                                <action name="minimal"/>
                                <action name="maximal" timeout="1" interval="2"
                                    start-delay="3" depth="4" role="whatever"
                                />
                                <action name="stonith_special"
                                    automatic="0" on_target="1"
                                />
                            </actions>
                        </resource-agent>
                    """
            ),
            ResourceAgentMetadataOcf1_1(
                self.agent_name,
                shortdesc=None,
                longdesc=None,
                parameters=[],
                actions=[
                    ResourceAgentActionOcf1_1(
                        name="minimal",
                        timeout=None,
                        interval=None,
                        role=None,
                        start_delay=None,
                        depth=None,
                        automatic=None,
                        on_target=None,
                    ),
                    ResourceAgentActionOcf1_1(
                        name="maximal",
                        timeout="1",
                        interval="2",
                        role="whatever",
                        start_delay="3",
                        depth="4",
                        automatic=None,
                        on_target=None,
                    ),
                    ResourceAgentActionOcf1_1(
                        name="stonith_special",
                        timeout=None,
                        interval=None,
                        role=None,
                        start_delay=None,
                        depth=None,
                        automatic="0",
                        on_target="1",
                    ),
                ],
            ),
        )



from pcs import settings
from pcs.lib.sbd import DEVICE_INITIALIZATION_OPTIONS_MAPPING

from pcs_test.tools.command_env.mock_runner import Call as RunnerCall


class SbdShortcuts:
    def __init__(self, calls):
        self.__calls = calls

    def list_watchdogs(
        self,
        output,
        name="runner.sbd.list_watchdogs",
        stderr="",
        returncode=0,
        instead=None,
        before=None,
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.sbd_binary, "query-watchdog"],
                stdout=output,
                stderr=stderr,
                returncode=returncode,
            ),
            before=before,
            instead=instead,
        )

    def initialize_devices(
        self,
        devices,
        options,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.sbd.initialize_devices",
    ):
        cmd = [settings.sbd_binary]
        for device in devices:
            cmd += ["-d", device]

        for opt, val in sorted(options.items()):
            cmd += [DEVICE_INITIALIZATION_OPTIONS_MAPPING[opt], str(val)]

        cmd.append("create")
        self.__calls.place(
            name,
            RunnerCall(
                cmd,
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )

    def get_device_info(
        self,
        device,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.sbd.device.list",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.sbd_binary, "-d", device, "list"],
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )

    def get_device_dump(
        self,
        device,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.sbd.device.dump",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.sbd_binary, "-d", device, "dump"],
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )

    def set_device_message(
        self,
        device,
        node,
        message,
        stdout="",
        stderr="",
        return_code=0,
        name="runner.sbd.device.set_message",
    ):
        self.__calls.place(
            name,
            RunnerCall(
                [settings.sbd_binary, "-d", device, "message", node, message],
                stdout=stdout,
                stderr=stderr,
                returncode=return_code,
            ),
        )
